#############################################################################
# neomail.pl - Provides a web interface to user mail spools                 #
# Copyright (C) 2000 Ernie Miller                                           #
#                                                                           #
# This program is free software; you can redistribute it and/or             #
# modify it under the terms of the GNU General Public License               #
# as published by the Free Software Foundation; either version 2            #
# of the License, or (at your option) any later version.                    #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software Foundation,   #
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.           #
#############################################################################

use Fcntl qw(:DEFAULT :flock);

################## GETFOLDERS ####################
sub getfolders {
   my (@folders, @userfolders);
   my @delfiles=();
   my $totalfoldersize = 0;
   my $filename;

   if ( $homedirfolders eq 'yes' ) {
      @folders = qw(INBOX saved_messages sent_mail neomail_trash);
      $totalfoldersize += ( -s "$folderdir/sent_mail" ) || 0;
      $totalfoldersize += ( -s "$folderdir/saved_messages" ) || 0;
      $totalfoldersize += ( -s "$folderdir/neomail_trash" ) || 0;
   } else {
      @folders = qw(INBOX SAVED TRASH SENT);
      $totalfoldersize += ( -s "$folderdir/SAVED" ) || 0;
      $totalfoldersize += ( -s "$folderdir/TRASH" ) || 0;
      $totalfoldersize += ( -s "$folderdir/SENT" ) || 0;
   }

   if (!(-d "$folderdir" && -e _)) {
	system("mkdir -p $folderdir");
   }
   opendir (FOLDERDIR, "$folderdir") || neomailerror("$lang_err{'couldnt_open'} $folderdir!");

   while (defined($filename = readdir(FOLDERDIR))) {

      ### files started with . are not folders
      ### they are neomail internal files (., .., dbm, search caches)
      if ( $filename=~/^\./ ) {
         if ( $filename=~/^\.(.*)\.db$/ ||
              $filename=~/^\.(.*)\.dir$/ ||
              $filename=~/^\.(.*)\.pag$/ ||
              ($filename=~/^\.(.*)\.cache$/ && $filename ne ".search.cache") ) {
            if ( ($1 ne $UIDNM{$menu_id}) && (! -f "$folderdir/$1") ){
               # dbm or cache whose folder doesn't exist
               ($filename =~ /^(.+)$/) && ($filename = $1); # bypass taint check
               push (@delfiles, "$folderdir/$filename");	
            }
         }
         next;
      }

      ### find all user folders
      if ( $homedirfolders eq 'yes' ) {
         unless ( ($filename eq 'saved_messages') ||
                  ($filename eq 'sent_mail') ||
                  ($filename eq 'neomail_trash') ||
                  ($filename eq '.') || ($filename eq '..')
                ) {
            push (@userfolders, $filename);
            $totalfoldersize += ( -s "$folderdir/$filename" );
         }
      } else {
         if ($filename =~ /^(.+)\.folder$/) {
            push (@userfolders, $1);
            $totalfoldersize += ( -s "$folderdir/$filename" );
         }
      }
   }

   closedir (FOLDERDIR) || neomailerror("$lang_err{'couldnt_close'} $folderdir!");

   push (@folders, sort(@userfolders));

   if ($#delfiles >= 0) {
      unlink(@delfiles);
   }

   if ($folderquota) {
      ($hitquota = 1) if ($totalfoldersize >= ($folderquota * 1024));
   }

   return \@folders;
}
################ END GETFOLDERS ##################

############### GET_SPOOLFILE_FOLDERDB ################
# this routine use lots of globals vars, ugly...
sub get_spoolfile_headerdb_and_set_uid_gid {
   my ($spoolfile, $headerdb, $username);

   $username=$UIDNM{$menu_id};
   if ($folder eq 'INBOX') {
      if ($homedirspools eq "yes") {
         $spoolfile = "$HOME{$menu_id}/$homedirspoolname";
      } elsif ($hashedmailspools eq "yes") {
         $username =~ /^(.)(.)/;
         my $firstchar = $1;
         my $secondchar = $2;
         $spoolfile = "$mailspooldir/$firstchar/$secondchar/$username";
      } else {
         $spoolfile = "$mailspooldir/$username";
	     # create system spool file /var/mail/xxxx
	     if ( ! -f "$spoolfile" ) {
		      open (F, ">>$spoolfile");
		      close(F);
		      chown($menu_id,$gid,$spoolfile);
		      chmod(0600, $spoolfile);
		 }
      }
      $headerdb="$folderdir/.$username";
   } elsif ( ($folder eq 'SAVED') || ($folder eq 'SENT') || 
      ($folder eq 'TRASH') || ( $homedirfolders eq 'yes' ) ) {
      $spoolfile = "$folderdir/$folder";
      $headerdb="$folderdir/.$folder";
   } else {
      $spoolfile = "$folderdir/$folder.folder";
      $headerdb= "$folderdir/.$folder.folder";
   }
   
   # chnage egid and euid if data is located in user dir
#   if ( ( ($homedirfolders eq 'yes') || ($homedirspools eq 'yes') ) && ($> == 0) ) {
#      $) = $gid;
#      $> = $menu_id;
#      umask(0077); # make sure only owner can read/write
#   }

   return($spoolfile, $headerdb);
}

############### GET_SPOOLFILE_FOLDERDB ################

################ DISPLAYHEADERS #####################
sub displayheaders {
   filtermessage();
   printheader();

   my ($bgcolor, $status, $message_size);
   my $newmessages = 0;
   my $escapedmessageid; # Used when creating link from subject line
   my @headers = @{&getheaders()};
   my $numheaders = $#headers + 1 || 1;
   my $page_total = $numheaders/$numberofheaders || 1;
   $page_total = int($page_total) + 1 if ($page_total != int($page_total));

   if (defined($DATA{"custompage"})) {
      my $pagenumber = $DATA{"custompage"};
      $pagenumber = 1 if ($pagenumber < 1);
      $pagenumber = $page_total if ($pagenumber > $page_total);
      $firstmessage = (($pagenumber-1)*$numberofheaders) + 1;
   }

### Perform verification of $firstmessage, make sure it's within bounds
   if ($firstmessage > ($#headers + 1)) {
      $firstmessage = $#headers - ($numberofheaders - 1);
   }
   if ($firstmessage < 1) {
      $firstmessage = 1;
   }
   my $lastmessage = $firstmessage + $numberofheaders - 1;
   if ($lastmessage > ($#headers + 1)) {
       $lastmessage = ($#headers + 1);
   }

   foreach my $messnum (0 .. $#headers) {
      unless (${$headers[$messnum]}{status} =~ /r/i) {
         $newmessages++;
      }
   }

   my $base_url = "$scripturl&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder";
   my $base_url_nokeyword = "$scripturl&amp;sort=$sort&amp;folder=$escapedfolder";

   my $page_nb;
   if ($#headers > 0) {
      $page_nb = ($firstmessage) * (($#headers + 1) / $numberofheaders) / ($#headers + 1);
      ($page_nb = int($page_nb) + 1) if ($page_nb != int($page_nb));
   } else {
      $page_nb = 1;
   }

   if ($total_size > 1048575){
      $total_size = int(($total_size/1048576)+0.5) . " MB";
   } elsif ($total_size > 1023) {
      $total_size =  int(($total_size/1024)+0.5) . " KB";
   } else {
      $total_size = $total_size . " B";
   }

   my $html = '';
   my $temphtml;
   open (VIEWFOLDER, "$neomaildir/templates/$lang/viewfolder.template") or
      neomailerror("$lang_err{'couldnt_open'} viewfolder.template!");
   while (<VIEWFOLDER>) {
      $html .= $_;
   }
   close (VIEWFOLDER);

   $html = applystyle($html);

   $temphtml = '</form>';
   $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/g;

### we don't keep keyword between folders,
### thus the keyword will be cleared when user change folder
   $temphtml = "<form method='POST' action='".$cgi_url."' name='FolderForm'>";
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
   $temphtml .= "<input type='hidden' name='action' value='displayheaders'>";
   $temphtml .= "<input type='hidden' name='firstmessage' value='".$firstmessage."'>";
   $html =~ s/\@\@\@STARTFOLDERFORM\@\@\@/$temphtml/;

   my %folderlabels;
   foreach (@validfolders) {
      my ($headerdb, $newmessages, $allmessages);

      if (defined $lang_folders{$_}) {
         $folderlabels{$_}=$lang_folders{$_};
      } else {
         $folderlabels{$_}=$_;
      }

      if ($_ eq 'INBOX') {
         $headerdb="$folderdir/.$UIDNM{$menu_id}";
      } else {
         $headerdb="$folderdir/.$_";
      }
      filelock("$headerdb.$dbm_ext", LOCK_SH);
      dbmopen (%HDB, $headerdb, undef);
      $allmessages=$HDB{'ALLMESSAGES'};
      $newmessages=$HDB{'NEWMESSAGES'};
      dbmclose(%HDB);
      filelock("$headerdb.$dbm_ext", LOCK_UN);

      if ( $newmessages ne "" && $allmessages ne "" ) {
         $folderlabels{$_}.= " ($newmessages/$allmessages)";
      }
   }

   $temphtml = "<select name='folder' onchange='JavaScript:document.FolderForm.submit();'>";
   foreach (@validfolders) {
      if ($_ eq $folder) {
            $temphtml .= "<option value='".$_."' selected>".$folderlabels{$_};
      } else {
            $temphtml .= "<option value='".$_."'>".$folderlabels{$_};
      }
   }
   $temphtml .= "</select>";
   $html =~ s/\@\@\@FOLDERPOPUP\@\@\@/$temphtml/;

   if (defined($headers[0])) {
      $temphtml = ($firstmessage) . " - " . ($lastmessage) . " $lang_text{'of'} " .
                  ($#headers + 1) . " $lang_text{'messages'} ";
      if ($newmessages) {
         $temphtml .= "($newmessages $lang_text{'unread'})";
      }
      $temphtml .= " - $total_size";
   } else {
      $temphtml = $lang_text{'nomessages'};
   }

   if ($hitquota) {
      $temphtml .= " [ $lang_text{'quota_hit'} ]";
   }
   $html =~ s/\@\@\@NUMBEROFMESSAGES\@\@\@/$temphtml/g;

   $temphtml = "<a href=\"$base_url&amp;action=composemessage&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/compose.gif\" border=\"0\" ALT=\"$lang_text{'composenew'}\"></a> ";
   $temphtml .= "<a href=\"$base_url_nokeyword&amp;action=displayheaders&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/refresh.gif\" border=\"0\" ALT=\"$lang_text{'refresh'}\"></a> ";
   $temphtml .= "<a href=\"$scripturl&amp;action=setprefs&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/prefs.gif\" border=\"0\" ALT=\"$lang_text{'userprefs'}\"></a> ";
   $temphtml .= "<a href=\"$scripturl&amp;action=editaddresses&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/addresses.gif\" border=\"0\" ALT=\"$lang_text{'addressbook'}\"></a> ";
   $temphtml .= "<a href=\"$scripturl&amp;action=editfolders&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/folder.gif\" border=\"0\" ALT=\"$lang_text{'folders'}\"></a> ";
   $temphtml .= "<a href=\"$scripturl&amp;action=editfilter&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/filtersetup.gif\" border=\"0\" ALT=\"$lang_text{'filterbook'}\"></a> &nbsp; &nbsp; ";
   $temphtml .= "<a href=\"$scripturl&amp;action=editpop3&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/pop3setup.gif\" border=\"0\" ALT=\"$lang_text{'pop3book'}\"></a> ";
   $temphtml .= "<a href=\"$scripturl&amp;action=retrpop3s&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/pop3.gif\" border=\"0\" ALT=\"$lang_text{'retr_pop3s'}\"></a> &nbsp; &nbsp; ";
   $temphtml .= "<a href=\"$base_url&amp;action=emptytrash&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/trash.gif\" border=\"0\" ALT=\"$lang_text{'emptytrash'}\"></a> ";
   $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

   $temphtml = "<form method='POST' action='".$cgi_url."'>";
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
   $temphtml .= "<input type='hidden' name='action' value='displayheaders'>";
   $temphtml .= "<input type='hidden' name='keyword' value='".$keyword."'>";
   $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
   $html =~ s/\@\@\@STARTPAGEFORM\@\@\@/$temphtml/g;
   
   my ($temphtml1, $temphtml2);

   if ($firstmessage != 1) {
      $temphtml1 = "<a href=\"$base_url&amp;action=displayheaders&amp;firstmessage=1\">";
      $temphtml1 .= "<img src=\"$image_url/first.gif\" align=\"absmiddle\" border=\"0\" alt=\"&lt;&lt;\"></a>";
   } else {
      $temphtml1 = "<img src=\"$image_url/first-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
   }

   if (($firstmessage - $numberofheaders) >= 1) {
      $temphtml1 .= "<a href=\"$base_url&amp;action=displayheaders&amp;firstmessage=" . ($firstmessage - $numberofheaders) . "\">";
      $temphtml1 .= "<img src=\"$image_url/left.gif\" align=\"absmiddle\" border=\"0\" alt=\"&lt;\"></a>";
   } else {
      $temphtml1 .= "<img src=\"$image_url/left-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
   }

   $html =~ s/\@\@\@LEFTPAGECONTROL\@\@\@/$temphtml1/g;

   if (($firstmessage + $numberofheaders) <= ($#headers + 1)) {
      $temphtml2 = "<a href=\"$base_url&amp;action=displayheaders&amp;firstmessage=" . ($firstmessage + $numberofheaders) . "\">";
      $temphtml2 .= "<img src=\"$image_url/right.gif\" align=\"absmiddle\" border=\"0\" alt=\"&gt;\"></a>";
   } else {
      $temphtml2 = "<img src=\"$image_url/right-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
   }

   if (($firstmessage + $numberofheaders) <= ($#headers +1) ) {
      $temphtml2 .= "<a href=\"$base_url&amp;action=displayheaders&amp;custompage=" . "$page_total\">";
      $temphtml2 .= "<img src=\"$image_url/last.gif\" align=\"absmiddle\" border=\"0\" alt=\"&gt;&gt;\"></a>";
   } else {
      $temphtml2 .= "<img src=\"$image_url/last-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
   }
   $html =~ s/\@\@\@RIGHTPAGECONTROL\@\@\@/$temphtml2/g;

   $temphtml = $temphtml1."&nbsp;&nbsp;"."[$lang_text{'page'} ";
   $temphtml .= "<input type='text' name='custompage' value='".$page_nb."' size='2'>";
   $temphtml .= " $lang_text{'of'} " . $page_total . ']'."&nbsp;&nbsp;".$temphtml2;
   $html =~ s/\@\@\@PAGECONTROL\@\@\@/$temphtml/g;

   $temphtml = "<form name='MoveForm' method='POST' action='".$cgi_url."' onsubmit=\"return confirm(".$lang_text{'moveconfirm'}.")\">";
   my @movefolders;
   foreach my $checkfolder (@validfolders) {
      unless ( ($checkfolder eq 'INBOX') || ($checkfolder eq $folder) ) {
         push (@movefolders, $checkfolder);
      }
   }
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='action' value='movemessage'>";
   $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
   $temphtml .= "<input type='hidden' name='keyword' value='".$keyword."'>";
   $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
   $temphtml .= "<input type='hidden' name='firstmessage' value='".$firstmessage."'>";
   $html =~ s/\@\@\@STARTMOVEFORM\@\@\@/$temphtml/g;
   
   if ( $homedirfolders eq 'yes' ) {
      $temphtml = "<select name='destination'>";
      foreach (@movefolders) {
         if ($_ eq 'neomail_trash') {
            $temphtml .= "<option value='".$_."' selected>".$folderlabels{$_};
         } else {
            $temphtml .= "<option value='".$_."'>".$folderlabels{$_};
         }
      }
      $temphtml .= "</select>";
   } else {
      $temphtml = "<select name='destination'>";
      foreach (@movefolders) {
         if ($_ eq 'TRASH') {
            $temphtml .= "<option value='".$_."' selected>".$folderlabels{$_};
         } else {
            $temphtml .= "<option value='".$_."'>".$folderlabels{$_};
         }
      }
      $temphtml .= "</select>";
   }
   $temphtml .= "<input type='submit' value='".$lang_text{'move'}."'>";
   $html =~ s/\@\@\@MOVECONTROLS\@\@\@/$temphtml/g;

   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;firstmessage=".
               ($firstmessage)."&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;sort=";
   $temphtml .= "status\"><IMG SRC=\"$image_url/new.gif\" border=\"0\" alt=\"$lang_sortlabels{'status'}\"></a>";
   $html =~ s/\@\@\@STATUS\@\@\@/$temphtml/g;
   
   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;firstmessage=".
               ($firstmessage)."&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;sort=";
   if ($sort eq "date") {
      $temphtml .= "date_rev\">$lang_text{'date'} <IMG SRC=\"$image_url/up.gif\" border=\"0\" alt=\"^\"></a>";
   } elsif ($sort eq "date_rev") {
      $temphtml .= "date\">$lang_text{'date'} <IMG SRC=\"$image_url/down.gif\" border=\"0\" alt=\"v\"></a>";
   } else {
      $temphtml .= "date\">$lang_text{'date'}</a>";
   }
   $html =~ s/\@\@\@DATE\@\@\@/$temphtml/g;
   
   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;firstmessage=".
                ($firstmessage)."&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;sort=";
   if ( ($folder eq 'SENT') || ($folder eq 'sent_mail') ) {
      if ($sort eq "sender") {
         $temphtml .= "sender_rev\">$lang_text{'recipient'} <IMG SRC=\"$image_url/down.gif\" border=\"0\" alt=\"v\"></a></B></td>";
      } elsif ($sort eq "sender_rev") {
         $temphtml .= "sender\">$lang_text{'recipient'} <IMG SRC=\"$image_url/up.gif\" border=\"0\" alt=\"^\"></a></B></td>";
      } else {
         $temphtml .= "sender\">$lang_text{'recipient'}</a>";
      }
   } else {
      if ($sort eq "sender") {
         $temphtml .= "sender_rev\">$lang_text{'sender'} <IMG SRC=\"$image_url/down.gif\" border=\"0\" alt=\"v\"></a>";
      } elsif ($sort eq "sender_rev") {
         $temphtml .= "sender\">$lang_text{'sender'} <IMG SRC=\"$image_url/up.gif\" border=\"0\" alt=\"^\"></a>";
      } else {
         $temphtml .= "sender\">$lang_text{'sender'}</a>";
      }
   }
   $html =~ s/\@\@\@SENDER\@\@\@/$temphtml/g;

   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;firstmessage=".
                ($firstmessage)."&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;sort=";
   if ($sort eq "subject") {
      $temphtml .= "subject_rev\">$lang_text{'subject'} <IMG SRC=\"$image_url/down.gif\" border=\"0\" alt=\"v\"></a>";
   } elsif ($sort eq "subject_rev") {
      $temphtml .= "subject\">$lang_text{'subject'} <IMG SRC=\"$image_url/up.gif\" border=\"0\" alt=\"^\"></a>";
   } else {
      $temphtml .= "subject\">$lang_text{'subject'}</a>";
   }
   $html =~ s/\@\@\@SUBJECT\@\@\@/$temphtml/g;

   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;firstmessage=".
                ($firstmessage)."&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;sort=";
   if ($sort eq "size") {
      $temphtml .= "size_rev\">$lang_text{'size'} <IMG SRC=\"$image_url/up.gif\" border=\"0\" alt=\"^\"></a>";
   } elsif ($sort eq "size_rev") {
      $temphtml .= "size\">$lang_text{'size'} <IMG SRC=\"$image_url/down.gif\" border=\"0\" alt=\"v\"></a>";
   } else {
      $temphtml .= "size\">$lang_text{'size'}</a>";
   }
   $html =~ s/\@\@\@SIZE\@\@\@/$temphtml/g;

   $temphtml = '';
   my ($boldon, $boldoff); # Used to control whether text is bold for new mails
   foreach my $messnum (($firstmessage - 1) .. ($lastmessage - 1)) {
### Stop when we're out of messages!
      last if !(defined($headers[$messnum]));

      (${$headers[$messnum]}{from} =~ s/^"?(.+?)"?\s*<(.*)>$/<a href="$scripturl&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;sessionid=$thissession&amp;composetype=sendto&amp;to=$2">$1<\/a>/) ||
      (${$headers[$messnum]}{from} =~ s/<?(.*@.*)>?\s+\((.+?)\)/<a href="$scripturl&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;sessionid=$thissession&amp;composetype=sendto&amp;to=$1">$2<\/a>/) ||
      (${$headers[$messnum]}{from} =~ s/(.+)/<a href="$scripturl&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;sessionid=$thissession&amp;composetype=sendto&amp;to=$1">$1<\/a>/);

      ${$headers[$messnum]}{subject} = str2html(${$headers[$messnum]}{subject});

### Make sure there's SOMETHING clickable for subject line
      unless (${$headers[$messnum]}{subject} =~ /[^\s]/) {
         ${$headers[$messnum]}{subject} = "N/A";
      }

      $escapedmessageid = &urlencode(${$headers[$messnum]}{message_id});

      if ( $messnum % 2 ) {
         $bgcolor = $style{"tablerow_light"};
      } else {
         $bgcolor = $style{"tablerow_dark"};
      }

      $message_size = ${$headers[$messnum]}{messagesize};
### Round message size and change to an appropriate unit for display
      if ($message_size > 1048575){
         $message_size = int(($message_size/1048576)+0.5) . "MB";
      } elsif ($message_size > 1023) {
         $message_size =  int(($message_size/1024)+0.5) . "KB";
      }

      $status = "<B>".($messnum+1)."</B> ";
### Choose status icons based on Status: line and type of encoding
      if ( ${$headers[$messnum]}{status} =~ /r/i ) {
         $boldon = '';
         $boldoff = '';
      } else {
         $status .= "<img src=\"$image_url/new.gif\" align=\"absmiddle\">";
         $boldon = "<B>";
         $boldoff = "</B>";
      }

      if ( (${$headers[$messnum]}{content_type} ne 'N/A') && !(${$headers[$messnum]}{content_type} =~ /^text/i) ) {
         $status .= "<img src=\"$image_url/attach.gif\" align=\"absmiddle\">";
      }

      $temphtml .= "<tr><td valign=\"middle\" width=\"50\" bgcolor=$bgcolor>$status&nbsp;</td>".
         "<td valign=\"middle\" width=\"150\" bgcolor=$bgcolor>$boldon<font size=-1>".
         ${$headers[$messnum]}{date}."</font>$boldoff</td>".
         "<td valign=\"middle\" width=\"150\" bgcolor=$bgcolor>$boldon".
         ${$headers[$messnum]}{from}."$boldoff</td>".
         "<td valign=\"middle\" width=\"350\" bgcolor=$bgcolor>".
         "<a href=\"$scripturl&amp;action=readmessage&amp;firstmessage=".
         ($firstmessage)."&amp;status=".
         ${$headers[$messnum]}{status}."&amp;folder=$escapedfolder&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;headers=".
         ($prefs{"headers"} || 'simple'). "&amp;message_id=".
         $escapedmessageid ."\">$boldon".
         ${$headers[$messnum]}{subject}."</a>$boldoff</td>".
         "<td valign=\"middle\" width=\"40\" bgcolor=$bgcolor>$boldon".
         $message_size . "$boldoff</td>".
         "<td align=\"center\" valign=\"middle\" width=\"50\" bgcolor=$bgcolor>".
         "<input type='checkbox' name='message_ids' value='".${$headers[$messnum]}{message_id}."'>".
         '</td></tr>';
   }
   $html =~ s/\@\@\@HEADERS\@\@\@/$temphtml/;

   $temphtml = "<form method='POST' action='".$cgi_url."'>";
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
   $temphtml .= "<input type='hidden' name='action' value='displayheaders'>";
   $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
   $html =~ s/\@\@\@STARTSEARCHFORM\@\@\@/$temphtml/g;

   $temphtml = "<b>$lang_text{search}&nbsp;&nbsp;</b>" .
                "<input type='text' name='keyword' value='".$keyword."' size='30'>";
   $html =~ s/\@\@\@SEARCH\@\@\@/$temphtml/g;

   print $html;

   printfooter();
}
############### END DISPLAYHEADERS ##################

################# READMESSAGE ####################

sub readmessage {
   printheader();
   my $messageid = $DATA{"message_id"};
   my $escapedmessageid = &urlencode($messageid);
   my %message = %{&getmessage($messageid)};
   my $headers = $DATA{"headers"} || 'simple';
   my $attmode = $DATA{"attmode"} || 'simple';

   if (%message) {
      my $html = '';
      my ($temphtml, $temphtml1, $temphtml2);
      open (READMESSAGE, "$neomaildir/templates/$lang/readmessage.template") or
         neomailerror("$lang_err{'couldnt_open'} readmessage.template!");
      while (<READMESSAGE>) {
         $html .= $_;
      }
      close (READMESSAGE);

      $html = applystyle($html);

      if (defined $lang_folders{$folder}) {
         $html =~ s/\@\@\@FOLDER\@\@\@/$lang_folders{$folder}/g;
      } else {
         $html =~ s/\@\@\@FOLDER\@\@\@/$folder/g;
      }

### these will hold web-ified headers
      my ($from, $replyto, $to, $cc, $subject, $body);
      $from = str2html($message{from} || '');
      $replyto = str2html($message{replyto} || '');
      $to = str2html($message{to} || '');
      $cc = str2html($message{cc} || '');
      $subject = str2html($message{subject} || '');

      $body = $message{"body"} || '';

      if ($message{contenttype} =~ /^text/i) {
         if ($message{encoding} =~ /^quoted-printable/i) {
            $body= decode_qp($body);
         } elsif ($message{encoding} =~ /^base64/i) {
            $body= decode_base64($body);
         }
      }
      if ($message{contenttype} =~ m#^text/html#i) { # convert into html table
         $body = html4nobase($body); 
         $body = html4mailto($body, $cgi_url, "step=mail&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;composetype=sendto");
         $body = html2table($body); 
      } else { 					     # body must be html or text
      # remove odds space or blank lines
         $body =~ s/(\r?\n){2,}/\n\n/g;
         $body =~ s/^\s+//;	
         $body =~ s/\n\s*$/\n/;
         $body = text2html($body);
      }

      my $base_url = "$scripturl&amp;firstmessage=" . ($firstmessage) .
                     "&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;message_id=$escapedmessageid";
      my $base_url_noid = "$scripturl&amp;firstmessage=" . ($firstmessage) .
                          "&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder";

##### Set up the message to go to after move.
      my $messageaftermove;
      if (defined($message{"next"})) {
         $messageaftermove = $message{"next"};
      } elsif (defined($message{"prev"})) {
         $messageaftermove = $message{"prev"};
      }

      $html =~ s/\@\@\@MESSAGETOTAL\@\@\@/$message{"total"}/g;

      $temphtml = "<a href=\"$base_url&amp;action=displayheaders\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a> &nbsp; &nbsp; ";
      $html =~ s/\@\@\@BACKTOLINK\@\@\@/$temphtml/g;

      $temphtml .= "<a href=\"$base_url&amp;action=composemessage&amp;composetype=reply\"><IMG SRC=\"$image_url/reply.gif\" border=\"0\" ALT=\"$lang_text{'reply'}\"></a> " .
      "<a href=\"$base_url&amp;action=composemessage&amp;composetype=replyall\"><IMG SRC=\"$image_url/replyall.gif\" border=\"0\" ALT=\"$lang_text{'replyall'}\"></a> " .
      "<a href=\"$base_url&amp;action=composemessage&amp;composetype=forward\"><IMG SRC=\"$image_url/forward.gif\" border=\"0\" ALT=\"$lang_text{'forward'}\"></a> &nbsp; &nbsp; ";
      $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

      if (defined($message{"prev"})) {
         $temphtml1 = "<a href=\"$base_url_noid&amp;action=readmessage&amp;message_id=$message{'prev'}&amp;headers=$headers&amp;attmode=$attmode\"><img src=\"$image_url/left.gif\" align=\"absmiddle\" border=\"0\" alt=\"&lt;&lt;\"></a>";
      } else {
         $temphtml1 = "<img src=\"$image_url/left-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
      }
      $html =~ s/\@\@\@LEFTMESSAGECONTROL\@\@\@/$temphtml1/g;

      if (defined($message{"next"})) {
         $temphtml2 = "<a href=\"$base_url_noid&amp;action=readmessage&amp;message_id=$message{'next'}&amp;headers=$headers&amp;attmode=$attmode\"><img src=\"$image_url/right.gif\" align=\"absmiddle\" border=\"0\" alt=\"&gt;&gt;\"></a>";
      } else {
         $temphtml2 = "<img src=\"$image_url/right-grey.gif\" align=\"absmiddle\" border=\"0\" alt=\"\">";
      }
      $html =~ s/\@\@\@RIGHTMESSAGECONTROL\@\@\@/$temphtml2/g;

      $temphtml = $temphtml1 . "  " . $message{"number"} . "  " . $temphtml2;
      $html =~ s/\@\@\@MESSAGECONTROL\@\@\@/$temphtml/g;

      $temphtml = "<form action=\"".$cgi_url."\" onsubmit=\"return confirm(".$lang_text{'moveconfirm'}.")\">";
	  my %folderlabels;
      my @movefolders;
      foreach my $checkfolder (@validfolders) {
		 if (defined $lang_folders{$checkfolder}) {
			$folderlabels{$checkfolder}=$lang_folders{$checkfolder};
	     } else {
		     $folderlabels{$checkfolder}=$checkfolder;
		 }
         unless ( ($checkfolder eq 'INBOX') || ($checkfolder eq $folder) ) {
            push (@movefolders, $checkfolder);
         }
      }
      $temphtml .= "<input type='hidden' name='step' value='mail'>";
      $temphtml .= "<input type='hidden' name='action' value='movemessage'>";
      $temphtml .= "<input type='hidden' name='firstmessage' value='".$firstmessage."'>";
      $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
      $temphtml .= "<input type='hidden' name='keyword' value='".$keyword."'>";
      $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
      $temphtml .= "<input type='hidden' name='message_ids' value='".$messageid."'>";
      if ($messageaftermove) {
         $temphtml .= "<input type='hidden' name='messageaftermove' value='1'>";
         $temphtml .= "<input type='hidden' name='message_id' value='".$messageaftermove."'>";
      }
      $html =~ s/\@\@\@STARTMOVEFORM\@\@\@/$temphtml/g;
   
      if ( $homedirfolders eq 'yes' ) {
         $temphtml = "<select name='destination'>";
         foreach (@movefolders) {
            if ($_ eq 'neomail_trash') {
               $temphtml .= "<option value='".$_."' selected>".$folderlabels{$_};
            } else {
               $temphtml .= "<option value='".$_."'>".$folderlabels{$_};
            }
         }
         $temphtml .= "</select>";
      } else {
         $temphtml = "<select name='destination'>";
         foreach (@movefolders) {
            if ($_ eq 'TRASH') {
               $temphtml .= "<option value='".$_."' selected>".$folderlabels{$_};
            } else {
               $temphtml .= "<option value='".$_."'>".$folderlabels{$_};
            }
         }
         $temphtml .= "</select>";
      }
      $temphtml .= "<input type=submit name='".$lang_text{'move'}."' value='".$lang_text{'move'}."'>";
      $html =~ s/\@\@\@MOVECONTROLS\@\@\@/$temphtml/g;

      if ($headers eq "all") {
         $message{"header"} = decode_mimewords($message{"header"});
         $message{"header"} = text2html($message{"header"});
         $message{"header"} =~ s/\n([-\w]+?:)/\n<B>$1<\/B>/g;
         $temphtml = $message{"header"};
      } else {
         $temphtml = "<B>$lang_text{'date'}:</B> $message{date}<BR>\n";

         $temphtml .= "<B>$lang_text{'from'}:</B> $from &nbsp;";
         my ($readname, $email);
         if  ( $message{from} =~ /^"?(.+?)"?\s*<(.*)>$/ ) {
            ($realname, $email)=($1, $2);
         } elsif ( $message{from} =~ /<?(.*@.*)>?\s+\((.+?)\)/ ) {
            ($realname, $email)=($2, $1);
         } else {
            ($realname, $email)=($from, $from);
         }
         $realname=&urlencode($realname);
         $email=&urlencode($email);
         $temphtml .= "<a href=\"$scripturl&amp;action=addaddress&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;realname=$realname&amp;email=$email&amp;message_id=$escapedmessageid\">".
                      "<IMG SRC=\"$image_url/imports.gif\" border=\"0\" ALT=\"$lang_text{'importadd'}\">".
                      "</a><BR>";

         if ($replyto) {
            $temphtml .= "<B>$lang_text{'replyto'}:</B> $replyto<BR>\n";
         }

         if ($to) {
            if ( length($to)>96 && $DATA{'receivers'} ne "all" ) {
              $to=substr($to,0,90)." ".
		  "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;attmode=$attmode&amp;receivers=all&amp;\">".
		  "<b>.....</b>"."</a>";
            }
            $temphtml .= "<B>$lang_text{'to'}:</B> $to<BR>\n";
         }

         if ($cc) {
            if ( length($cc)>96 && $DATA{'receivers'} ne "all" ) {
              $cc=substr($cc,0,90)." ".
		  "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;attmode=$attmode&amp;receivers=all&amp;\">".
		  "<b>.....</b>"."</a>";
            }
            $temphtml .= "<B>$lang_text{'cc'}:</B> $cc<BR>\n";
         }

         if ($subject) {
            $temphtml .= "<B>$lang_text{'subject'}:</B> $subject\n";
         }
      }

      $html =~ s/\@\@\@HEADERS\@\@\@/$temphtml/g;

#      if ($headers eq "all") {
#         $temphtml = "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;attmode=$attmode&amp;headers=simple\">$lang_text{'simplehead'}</a>";
#      } else {
#         $temphtml = "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;attmode=$attmode&amp;headers=all\">$lang_text{'allhead'}</a>";
#      }
      $temphtml = '';
      $html =~ s/\@\@\@HEADERSTOGGLE\@\@\@/$temphtml/g;

      if ( $#{$message{attachment}} >=0 ) {
         if ($attmode eq "all") {
            $temphtml = "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;headers=$headers&amp;attmode=simple\">$lang_text{'simpleattmode'}</a>";
         } else {
            $temphtml = "<a href=\"$base_url&amp;action=readmessage&amp;message_id=$escapedmessageid&amp;headers=$headers&amp;attmode=all\">$lang_text{'allattmode'}</a>";
         }
      } else {
         $temphtml="&nbsp";
      }
      $html =~ s/\@\@\@ATTMODETOGGLE\@\@\@/$temphtml/g;

      $temphtml=$body;
      if ( $attmode eq 'all' ) {
         $temphtml .= "<hr>" if ( $#{$message{attachment}}>=0 );
      } else {
         $temphtml="" if ( $message{contenttype} =~ /^multipart/i );
      }

      foreach my $attnumber (0 .. $#{$message{attachment}}) {
         next unless (defined(%{$message{attachment}[$attnumber]}));

         if ( $attmode eq 'all' ) {
            if ( ${$message{attachment}[$attnumber]}{filename}=~
							/\.(jpg|jpeg|gif|png)$/i) {
               $temphtml .= image_att2table($message{attachment}, $attnumber, $escapedmessageid);
            } else {
               $temphtml .= misc_att2table($message{attachment}, $attnumber, $escapedmessageid);
            }

         } else {	# attmode==simple

            # handle case to skip to next text/html attachment
            if ( defined(%{$message{attachment}[$attnumber+1]}) &&
                 (${$message{attachment}[$attnumber+1]}{boundary} eq 
		  ${$message{attachment}[$attnumber]}{boundary}) ) {

               # skip to next text/html attachment in the same alternative group
               if ( (${$message{attachment}[$attnumber]}{subtype} =~ /alternative/i) &&
                 (${$message{attachment}[$attnumber+1]}{subtype} =~ /alternative/i) &&
                 (${$message{attachment}[$attnumber+1]}{contenttype} =~ /^text/i) ) {
                  next;
               }
               # skip to next attachment if this=unknow.txt and next=unknow.html
               if ( (${$message{attachment}[$attnumber]}{contenttype}=~ /^text\/plain/i ) &&
                    (${$message{attachment}[$attnumber]}{filename}=~ /^Unknown\./ ) &&
                    (${$message{attachment}[$attnumber+1]}{contenttype} =~ /^text\/html/i)  &&
                    (${$message{attachment}[$attnumber+1]}{filename}=~ /^Unknown\./ ) ) {
                  next;
               }
            }

            # handle display of attachments in simple mode
            if ( ${$message{attachment}[$attnumber]}{contenttype}=~ /^text\/html/i ) {
               if ( ${$message{attachment}[$attnumber]}{filename}=~ /^Unknown\.msg/ ) {

                  my ($header, $body)=
			split(/\n\r*\n/, ${$message{attachment}[$attnumber]}{contents}, 2);
                  $header=simpleheader($header);
                  #set bgcolor for message/rfc822 header
                  $header=~s/#dddddd/$style{"window_dark"}/ig;	
                  ${$message{attachment}[$attnumber]}{contents}=$header."\n\n".$body;

                  $temphtml .= html_att2table($message{attachment}, $attnumber, $escapedmessageid);
               } elsif ( ${$message{attachment}[$attnumber]}{filename}=~ /^Unknown\./ ) {
                  $temphtml .= html_att2table($message{attachment}, $attnumber, $escapedmessageid);
               } else {
                  $temphtml .= misc_att2table($message{attachment}, $attnumber, $escapedmessageid);
               }
            } elsif ( ${$message{attachment}[$attnumber]}{contenttype}=~ /^text/i ) {
               if ( ${$message{attachment}[$attnumber]}{filename}=~ /^Unknown\./ ) {
                  $temphtml .= text_att2table($message{attachment}, $attnumber);
               } else {
                  $temphtml .= misc_att2table($message{attachment}, $attnumber, $escapedmessageid);
               }
            } elsif ( ${$message{attachment}[$attnumber]}{filename}=~ /\.(jpg|jpeg|gif|png)$/i) {
               # show image only if it is not linked
               if ( ($decodedhtml !~ /\Q${$message{attachment}[$attnumber]}{id}\E/) &&
                    ($decodedhtml !~ /\Q${$message{attachment}[$attnumber]}{location}\E/) ) {
                  $temphtml .= image_att2table($message{attachment}, $attnumber, $escapedmessageid);
               }
            } else {
               $temphtml .= misc_att2table($message{attachment}, $attnumber, $escapedmessageid);
            }

         }
      }

      # enable download the whole message block in attmode==all
      if ( $attmode eq 'all' ) {
         $temphtml .= qq|<table width="100%" border="0" align="center" cellpadding="2"><tr><td align=right>|.
                      qq|<a href="$cgi_url/Unknow.msg?step=mail&amp;action=viewattachment&amp;message_id=$escapedmessageid&amp;folder=$escapedfolder&amp;attachment_number=all">$lang_text{'download'}</a>|.
                      qq|</td></tr></table>|;
      }

      $html =~ s/\@\@\@BODY\@\@\@/$temphtml/g;
      print $html;

   } else {
      $messageid = str2html($messageid);
      print "What the heck? Message $messageid seems to be gone!";
   }
   printfooter();

### fork a child to do the status update and headerdb update
   unless ($message{status} =~ /r/i) {
      $|=1; 				# flush all output
      $SIG{CHLD} = sub { wait };	# handle zombie
      if ( fork() == 0 ) {		# child
         close(STDOUT);
         close(STDIN);
         updatestatus($messageid,"R");
         exit;
      }
   }
}


sub html_att2table {
   my ($r_attachments, $attnumber, $escapedmessageid)=@_;

   my $r_attachment=${$r_attachments}[$attnumber];
   my $temphtml;

   if (${$r_attachment}{encoding} =~ /^quoted-printable/i) {
      $temphtml = decode_qp(${$r_attachment}{contents});
   } elsif (${$r_attachment}{encoding} =~ /^base64/i) {
      $temphtml = decode_base64(${$r_attachment}{contents});
   } else {
      $temphtml = ${$r_attachment}{contents};
   }

   $decodedhtml = $temphtml;	# store decoded html in global, used by others 

   $temphtml = html4nobase($temphtml);
   $temphtml = html4attachments($temphtml, $r_attachments, $cgi_url, "step=mail&amp;action=viewattachment&amp;message_id=$escapedmessageid&amp;folder=$escapedfolder");
   $temphtml = html4mailto($temphtml, $cgi_url, "step=mail&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;composetype=sendto");
   $temphtml = html2table($temphtml);

   return($temphtml);
}

sub text_att2table {
   my ($r_attachments, $attnumber)=@_;

   my $r_attachment=${$r_attachments}[$attnumber];
   my $temptext;

   if (${$r_attachment}{encoding} =~ /^quoted-printable/i) {
      $temptext = decode_qp(${$r_attachment}{contents});
   } elsif (${$r_attachment}{encoding} =~ /^base64/i) {
      $temptext = decode_base64(${$r_attachment}{contents});
   } else {
      $temptext = ${$r_attachment}{contents};
   }
 
   # remove odds space or blank lines
   $temptext =~ s/(\r?\n){2,}/\n\n/g;
   $temptext =~ s/^\s+//;	
   $temptext =~ s/\n\s*$/\n/;
   return(text2html($temptext). "<BR>");
}

sub image_att2table {
   my ($r_attachments, $attnumber, $escapedmessageid)=@_;

   my $r_attachment=${$r_attachments}[$attnumber];
   my $escapedfilename = &urlencode(${$r_attachment}{filename});
   my $attlen=lenstr(length(${$r_attachment}{contents}));

   my $temphtml .= qq|<table border="0" align="center" cellpadding="2">|.
                   qq|<tr><td valign="middle" bgcolor=$style{"attachment_dark"} align="center">|.
                   qq|$lang_text{'attachment'} $attnumber: ${$r_attachment}{filename} &nbsp;($attlen)|.
                   qq|</td></tr><td valign="middle" bgcolor=$style{"attachment_light"} align="center">|.
                   qq|<IMG BORDER="0" SRC="$cgi_url/$escapedfilename?step=mail&amp;action=viewattachment&amp;message_id=$escapedmessageid&amp;folder=$escapedfolder&amp;attachment_number=$attnumber">|.
                   qq|</td></tr></table>|;
   return($temphtml);
}

sub misc_att2table {
   my ($r_attachments, $attnumber, $escapedmessageid)=@_;

   my $r_attachment=${$r_attachments}[$attnumber];
   my $escapedfilename = &urlencode(${$r_attachment}{filename});
   my $attlen=lenstr(length(${$r_attachment}{contents}));

   my $temphtml .= qq|<table border="0" width="40%" align="center" cellpadding="2">|.
                   qq|<tr><td nowrap colspan="2" valign="middle" bgcolor=$style{"attachment_dark"} align="center">|.
                   qq|$lang_text{'attachment'} $attnumber: ${$r_attachment}{filename}&nbsp;($attlen)|.
                   qq|</td></tr>|.
                   qq|<tr><td nowrap valign="middle" bgcolor= $style{"attachment_light"} align="center">|.
                   qq|$lang_text{'type'}: ${$r_attachment}{contenttype}<br>|.
                   qq|$lang_text{'encoding'}: ${$r_attachment}{encoding}|.
                   qq|</td><td nowrap width="10%" valign="middle" bgcolor= $style{"attachment_light"} align="center">|.
                   qq|<a href="$cgi_url/$escapedfilename?step=mail&amp;action=viewattachment&amp;message_id=$escapedmessageid&amp;folder=$escapedfolder&amp;attachment_number=$attnumber">$lang_text{'download'}</a>|.
                   qq|</td></tr></table>|;
   return($temphtml);
}

sub lenstr {
   my $len=$_[0];

   if ($len > 1048575){
      $len = int(($len/1048576)+0.5) . "MB";
   } elsif ($len > 1023) {
      $len =  int(($len/1024)+0.5) . "KB";
   } else {
      $len = $len . "byte";
   }
   return ($len);
}

############### END READMESSAGE ##################

############### COMPOSEMESSAGE ###################
sub composemessage {
   no strict 'refs';
   my $html = '';
   my $temphtml;
   my @attlist;
   open (COMPOSEMESSAGE, "$neomaildir/templates/$lang/composemessage.template") or
      neomailerror("$lang_err{'couldnt_open'} composemessage.template!");
   while (<COMPOSEMESSAGE>) {
      $html .= $_;
   }
   close (COMPOSEMESSAGE);

   $html = applystyle($html);
   
   if (defined($DATA{$lang_text{'add'}})) {
      @attlist = @{&getattlist()};
   } elsif ( !(defined($DATA{$lang_text{'add'}})) ) {
      deleteattachments();
   }

   my $messageid = $DATA{"message_id"};
   my %message = ();
   my $attnumber = 0;
   my $from = '';
   if ($prefs{"fromname"}) {
      $from = $prefs{"fromname"} . "@" . $prefs{domainname}; 
   } else {
      $from = $UIDNM{$menu_id}."@".$prefs{domainname}; # create from: address
   } 
   if ($prefs{"realname"}) {
      my $realname = $prefs{"realname"};
      $from =~ s/^(.+)$/$realname <$1>/;
   }
   my $escapedfrom = str2html($from);
   my $to = '';
   my $cc = '';
   my $bcc = '';
   my $subject = '';
   my $body = '';
   my $composetype = $DATA{"composetype"};

   if ($composetype) {
      $to = $DATA{"to"} || '';
      $cc = $DATA{"cc"} || '';
      $bcc = $DATA{"bcc"} || '';
      $subject = $DATA{"subject"} || '';
      $body = $DATA{"body"} || '';

      if (($composetype eq "reply") || ($composetype eq "replyall") ||
          ($composetype eq "forward") ) {
         %message = %{&getmessage($messageid)};

### Handle mail programs that send the body of a message quoted-printable
         if ( ($message{contenttype} =~ /^text/i) &&
            ($message{encoding} =~ /^quoted-printable/i) ) {
            $message{"body"} = decode_qp($message{"body"});
         }
         $body = $message{"body"} || '';

### If the first attachment is text, assume it's the body of a message
### in multi-part format
         if (($message{contenttype} =~ /^multipart/i) &&
            (defined(${$message{attachment}[0]}{contenttype})) &&
            (${$message{attachment}[0]}{contenttype} =~ /^text\/plain/i)) {
            if (${$message{attachment}[0]}{encoding} =~ /^quoted-printable/i) {
               ${$message{attachment}[0]}{contents} =
               decode_qp(${$message{attachment}[0]}{contents});
            } elsif (${$message{attachment}[$attnumber]}{encoding} =~ /^base64/i) {
               	 ${$message{attachment}[$attnumber]}{contents} = decode_base64(${$message{attachment}[$attnumber]}{contents});
            }
            $body = ${$message{attachment}[0]}{contents};
            shift @{$message{attachment}};
         }

# Handle the messages generated if sendmail is set up to send MIME error reports
         if ($message{contenttype} =~ /^multipart\/report/i) {
            foreach my $attnumber (0 .. $#{$message{attachment}}) {
               if (defined(${$message{attachment}[$attnumber]}{contents})) {
                  $body .= ${$message{attachment}[$attnumber]}{contents};
                  shift @{$message{attachment}};
               }
            }
         }
      }

# remove odds space or blank lines
      $body =~ s/(\r?\n){2,}/\n\n/g;
      $body =~ s/^\s+//;	
      $body =~ s/\s+$//;

      if (($composetype eq "reply") || ($composetype eq "replyall")) {
         $subject = $message{"subject"} || '';
         $subject = "Re: " . $subject unless ($subject =~ /^re:/i);
         if (defined($message{"replyto"})) {
            $to = $message{"replyto"} || '';
         } else {
            $to = $message{"from"} || '';
         }
         if ($composetype eq "replyall") {
            $to .= "," . $message{"to"} if (defined($message{"to"}));
            $to .= "," . $message{"cc"} if (defined($message{"cc"}));
         }

         $body =~ s/\n/\n\> /g;
         $body = "> " . $body . "\n\n";
      }

      if ($composetype eq "forward") {
         if (defined(${$message{attachment}[0]}{header})) {
            foreach my $attnumber (0 .. $#{$message{attachment}}) {
	       ($attnumber =~ /^(.+)$/) && ($attnumber = $1);   # bypass taint check
               open (ATTFILE, ">$neomaildir$UIDNM{$menu_id}-att$attnumber");
               print ATTFILE ${$message{attachment}[$attnumber]}{header}, "\n\n", ${$message{attachment}[$attnumber]}{contents};
               close ATTFILE;
            }
            @attlist = @{&getattlist()};
         }
         $subject = $message{"subject"} || '';
         $subject = "Fw: " . $subject unless ($subject =~ /^fw:/i);
         $body = "\n\n------------- Forwarded message follows -------------\n$body";
      }

   }
   if ( (defined($prefs{"signature"})) && ($composetype ne 'continue') ) {
      $body .= "\n\n".$prefs{"signature"};
   }
   printheader();
   
   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;folder=$escapedfolder&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;firstmessage=$firstmessage\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a>";
   $html =~ s/\@\@\@BACKTOFOLDER\@\@\@/$temphtml/g;

   $temphtml = '<FORM METHOD="POST"  ENCTYPE="multipart/form-data" NAME="composeform">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="sendmessage">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="composetype" VALUE="continue">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="keyword" VALUE="'.$keyword.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   $html =~ s/\@\@\@STARTCOMPOSEFORM\@\@\@/$temphtml/g;

   $html =~ s/\@\@\@ESCAPEDFROM\@\@\@/$escapedfrom/g;

   $temphtml = '<INPUT TYPE="text" NAME="to" SIZE=70 value="'.str2html($to).'">';
   $html =~ s/\@\@\@TOFIELD\@\@\@/$temphtml/g;

   $temphtml = '<INPUT TYPE="text" NAME="cc" SIZE=70 value="'.str2html($cc).'">';
   $html =~ s/\@\@\@CCFIELD\@\@\@/$temphtml/g;
          
   $temphtml = '<INPUT TYPE="text" NAME="bcc" SIZE=70 value="'.str2html($bcc).'">';
   $html =~ s/\@\@\@BCCFIELD\@\@\@/$temphtml/g;
 
   $temphtml = '<INPUT TYPE="text" NAME="replyto" SIZE=70 value='.$prefs{"replyto"}.'>';
   $html =~ s/\@\@\@REPLYTOFIELD\@\@\@/$temphtml/g;
   
   $temphtml = '';
   foreach my $filename (@attlist) {
      $temphtml .= "$filename<BR>";
   }
   if ( $savedattsize ) {
      $temphtml .= "<em>" . int($savedattsize/1024) . "KB";
      if ( $attlimit ) {
         $temphtml .= " $lang_text{'of'} $attlimit MB";
      }
      $temphtml .= "</em><BR>";
   }
   $temphtml .= '<INPUT TYPE="file" NAME="attachment" SIZE=60 TABINDEX="-1">';
   $temphtml .= '<INPUT TYPE="submit" NAME="'.$lang_text{'add'}.'" VALUE="'.$lang_text{'add'}.'" TABINDEX="-1">';
   $html =~ s/\@\@\@ATTACHMENTFIELD\@\@\@/$temphtml/g;

   $temphtml = '<INPUT TYPE="text" NAME="subject" SIZE=70 value='.$subject.'>';
   $html =~ s/\@\@\@SUBJECTFIELD\@\@\@/$temphtml/g;

   $temphtml = '<TEXTAREA NAME="body" ROWS=20 COLS=78 WRAP="hard">'.$body.'</TEXTAREA>';
   $html =~ s/\@\@\@BODYAREA\@\@\@/$temphtml/g;

   $temphtml = '<INPUT TYPE="submit" NAME="'.$lang_text{'send'}.'" VALUE="'.$lang_text{'send'}.'">';
   $html =~ s/\@\@\@SENDBUTTON\@\@\@/$temphtml/g;

   $temphtml = "</form>";
   $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/g;

   $temphtml = '<FORM METHOD="POST" ACTION="'.$cgi_url.'">';
   if ($DATA{"message_id"}) {
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="readmessage">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="keyword" VALUE="'.$keyword.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="headers" VALUE="'.($prefs{"headers"} || 'simple').'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$DATA{"message_id"}.'">';
   } else {
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="displayheaders">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="keyword" VALUE="'.$keyword.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   }
   $html =~ s/\@\@\@STARTCANCELFORM\@\@\@/$temphtml/g;

   $temphtml = '<INPUT TYPE="submit" NAME="'.$lang_text{'cancel'}.'" VALUE="'.$lang_text{'cancel'}.'">';
   $html =~ s/\@\@\@CANCELBUTTON\@\@\@/$temphtml/g;

   print $html;

   printfooter();
}
############# END COMPOSEMESSAGE #################

############### SENDMESSAGE ######################
sub sendmessage {
   no strict 'refs';

   if (defined($DATA{$lang_text{'add'}})) {
      composemessage();
   } else {
### Add a header that will allow SENT folder to function correctly
      my $localtime = scalar(localtime);
      my $messagecontents = "From $UIDNM{$menu_id} $localtime\n";
      my $date = localtime();
      my @datearray = split(/ +/, $date);
      $date = "$datearray[0], $datearray[2] $datearray[1] $datearray[4] $datearray[3] $timeoffset";
      my $from;
      my $realname = $prefs{"realname"} || '';
      if($prefs{"fromname"}) {
         # Create from: address for when "fromname" is defined
         $from = $prefs{"fromname"} . "@" . $prefs{domainname};
      } else {
         # Create from: address for when "fromname" is not defined
         $from = $UIDNM{$menu_id} . "@" . $prefs{domainname};
      }
      $from =~ s/[\||'|"|`]/ /g;  # Get rid of shell escape attempts
      $realname =~ s/[\||'|"|`]/ /g;  # Get rid of shell escape attempts
      ($realname =~ /^(.+)$/) && ($realname = '"'.$1.'"');
      ($from =~ /^(.+)$/) && ($from = $1);

      my $savedatts = ''; # Will buffer saved attachments
      my $boundary = "----=NEOMAIL_ATT_" . rand();
      my $to = $DATA{"to"};
      my $cc = $DATA{"cc"};
      my $bcc = $DATA{"bcc"};
      my $subject = $DATA{"subject"};
      my $body = $DATA{"body"};
      $body =~ s/\r//g;  # strip ^M characters from message. How annoying!
      my $attachment = $DATA{"attachment"};
      if ( $attachment ) {
         getattlist();
         if ( ($attlimit) && ( ( $savedattsize + (-s $attachment) ) > ($attlimit * 1048576) ) ) {
            neomailerror ("$lang_err{'att_overlimit'} $attlimit MB!");
         }
      }
      my $attname = $attachment;
### Convert :: back to the ' like it should be.
      $attname =~ s/::/'/g;
### Trim the path info from the filename
      $attname =~ s/^.*\\//;
      $attname =~ s/^.*\///;
      $attname =~ s/^.*://;

      open (SENDMAIL, "|" . $sendmail . " -oem -oi -F '$realname' -f '$from' -t 1>&2") or
         neomailerror("$lang_err{'couldnt_open'} $sendmail!");
      print SENDMAIL "From: $realname <$from>\n";
      $messagecontents .= "From: $realname <$from>\n";
      print SENDMAIL "To: $to\n";
      $messagecontents .= "To: $to\n";
      if ($cc) {
         print SENDMAIL "CC: $cc\n";
         $messagecontents .= "CC: $cc\n";
      }
      if ($bcc) {
         print SENDMAIL "Bcc: $bcc\n";
         $messagecontents .= "Bcc: $bcc\n";
      }
      if ($prefs{"replyto"}) {
         print SENDMAIL "Reply-To: ",$prefs{"replyto"},"\n";
         $messagecontents .= "Reply-To: ".$prefs{"replyto"}."\n";
      }
      print SENDMAIL "Subject: $subject\n";
      $messagecontents .= "Subject: $subject\n";
      print SENDMAIL "X-Mailer: NeoMail $version\n";
      $messagecontents .= "X-Mailer: NeoMail $version\n";
      print SENDMAIL "X-IPAddress: $userip\n";
      $messagecontents .= "X-IPAddress: $userip\n";

      my $messageid="<NeoMail-saved-".rand().">";
      $messagecontents .= "Message-Id: $messageid\nDate: $date\nStatus: R\n";

      print SENDMAIL "MIME-Version: 1.0\n";
      $messagecontents .= "MIME-Version: 1.0\n";

      opendir (NEOMAILDIR, "$neomaildir") or
         neomailerror("$lang_err{'couldnt_open'} $neomaildir!");
      while (defined(my $currentfile = readdir(NEOMAILDIR))) {
         if ($currentfile =~ /^($UIDNM{$menu_id}-att\d+)$/) {
            $currentfile = $1;
            open (ATTFILE, "$neomaildir$currentfile");
            $/ = undef; # Force a single input to grab the whole spool
            $savedatts .= "\n--$boundary\n" . <ATTFILE>;
            $/ = "\n";
            close (ATTFILE);
         }
      }
      closedir (NEOMAILDIR);

      my $contenttype="";
      if ($attachment || $savedatts) {
         $contenttype="multipart/mixed;";
         print SENDMAIL "Content-Type: multipart/mixed;\n";
         print SENDMAIL "\tboundary=\"$boundary\"\n\n";
         print SENDMAIL "This is a multi-part message in MIME format.\n\n";
         print SENDMAIL "--$boundary\n";
         print SENDMAIL "Content-Type: text/plain; charset=$lang_charset\n\n";
         print SENDMAIL $body;
         $body =~ s/^From />From /gm;
         print SENDMAIL "\n$savedatts\n";

         $messagecontents .= "Content-Type: multipart/mixed;\n".
                             "\tboundary=\"$boundary\"\n\nThis is a multi-part message in MIME format.\n\n".
                             "--$boundary\nContent-Type: text/plain; charset=$lang_charset\n\n$body\n$savedatts\n";
         if ($attachment) {
            my $attcontents = '';
            $/ = undef; # Force a single input to grab the whole spool
            $attcontents = <$attachment>;
            $/ = "\n";
            $attcontents = encode_base64($attcontents);
            my $content_type;
            if (defined(uploadInfo($attachment))) {
               $content_type = ${uploadInfo($attachment)}{'Content-Type'} || 'application/octet-stream';
            } else {
               $content_type = 'application/octet-stream';
            }
            print SENDMAIL "--$boundary\nContent-Type: ", $content_type,";\n";
            print SENDMAIL "\tname=\"$attname\"\nContent-Transfer-Encoding: base64\n\n";
            print SENDMAIL "$attcontents\n";
            $messagecontents .= "--$boundary\nContent-Type: $content_type;\n".
                             "\tname=\"$attname\"\nContent-Transfer-Encoding: base64\n\n".
                             "$attcontents\n";
         }
         print SENDMAIL "--$boundary--";
         $messagecontents .= "--$boundary--\n\n";
      } else {
         print SENDMAIL "Content-Type: text/plain; charset=$lang_charset\n\n$body\n";
         $body =~ s/^From />From /gm;
         $messagecontents .= "Content-Type: text/plain; charset=$lang_charset\n\n$body\n\n";
      }

      close(SENDMAIL) or senderror();
      
      deleteattachments();
      
      my $sentfolder;
      if ( $homedirfolders eq 'yes' ) {
         $sentfolder = 'sent_mail';
      } else {
         $sentfolder = 'SENT';
      }

#      if ( ( ($homedirfolders eq 'yes') || ($homedirspools eq 'yes') ) && ($> == 0) ) {
#        $) = $gid;
#         $> = $menu_id;
#         umask(0077); # make sure only owner can read/write
#      }

      unless ($hitquota) {
         unless (filelock("$folderdir/$sentfolder", LOCK_EX|LOCK_NB)) {
            neomailerror("$lang_err{'couldnt_lock'} $sentfolder!");
         }
         update_headerdb("$folderdir/.$sentfolder", "$folderdir/$sentfolder");
         open (SENT, ">>$folderdir/$sentfolder") or
            neomailerror ("$lang_err{'couldnt_open'} $sentfolder!");

         my @attr;
         my @datearray = split(/\s+/, $date);
         if ($datearray[0] =~ /[A-Za-z,]/) {
            shift @datearray; # Get rid of the day of the week
         }

         $attr[$_OFFSET]=tell(SENT);
         $attr[$_TO]=$to;
         $attr[$_FROM]=(split(/,/, $to))[0];	# since this is sent folder
### we store localtime on dbm, so day offset is not needed
         $attr[$_DATE]="$month{$datearray[1]}/$datearray[0]/$datearray[2] $datearray[3]";
         $attr[$_SUBJECT]=$subject;
         $attr[$_CONTENT_TYPE]=$contenttype;
         $attr[$_STATUS]="R";
         $attr[$_SIZE]=length($messagecontents)+length("\n");

         print SENT $messagecontents,"\n";
         close (SENT) or neomailerror("$lang_err{'couldnt_close'} $sentfolder!");

         my %HDB;
         filelock("$folderdir/.$sentfolder.$dbm_ext", LOCK_EX);
         dbmopen(%HDB, "$folderdir/.$sentfolder", 0600);
         $HDB{$messageid}=join('@@@', @attr);
         $HDB{'ALLMESSAGES'}++;
         $HDB{'METAINFO'}=metainfo("$folderdir/$sentfolder");
         dbmclose(%HDB);
         filelock("$folderdir/.$sentfolder.$dbm_ext", LOCK_UN);

         filelock("$folderdir/$sentfolder", LOCK_UN);
      }
      displayheaders();
   }
}

sub senderror {
   my $html = '';
   my $temphtml;
   open (SENDERROR, "$neomaildir/templates/$lang/senderror.template") or
      neomailerror("$lang_err{'couldnt_open'} senderror.template!");
   while (<SENDERROR>) {
      $html .= $_;
   }
   close (SENDERROR);

   $html = applystyle($html);

   printheader();

   print $html;

   printfooter();
   exit 0;
}
############## END SENDMESSAGE ###################

################ VIEWATTACHMENT ##################
sub viewattachment {
   my $spoolfile;
   my $headerdb;
   my $spoolhandle=FileHandle->new();

   my $messageid = $DATA{"message_id"};
   my $attnum = $DATA{"attachment_number"};

   ($spoolfile, $headerdb)=get_spoolfile_headerdb_and_set_uid_gid();

   unless(filelock($spoolfile, LOCK_SH|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_locksh'} $spoolfile!");
   }
   update_headerdb($headerdb, $spoolfile);
   open($spoolhandle, "$spoolfile");
   my $r_block= get_message_block($messageid, $headerdb, $spoolhandle);
   close($spoolhandle);
   filelock($spoolfile, LOCK_UN);
   chown($menu_id,$gid,$spoolfile);

#   if ( ! defined(${$r_block}) ) {
#      printheader();
#      print "What the heck? Message seems to be gone!";
#      printfooter();
#      return;
#   }

   if ( $attnum eq 'all' ) {
      # return whole msg as an message/rfc822 object
      my $length = length(${$r_block});
      print qq|Content-Length: $length\n|.
            qq|Content-Transfer-Coding: binary\n|.
            qq|Connection: close\n|.
            qq|Content-Type: message/rfc822; name="Unknown.msg"\n|.
            qq|Content-Disposition: attachment; filename="Unknown.msg"\n|.
            qq|\n|.
            qq|${$r_block}|;

   } else {
      # return a specific attachment

      my ($header, $body, $r_attachments)=parse_rfc822block($r_block);
      $r_attachment=${$r_attachments}[$attnum];

      if ($r_attachment) {
         my $content;

         if (${$r_attachment}{encoding} =~ /^base64$/i) {
            $content = decode_base64(${$r_attachment}{contents});
         } elsif (${$r_attachment}{encoding} =~ /^quoted-printable$/i) {
            $content = decode_qp(${$r_attachment}{contents});
#         } elsif (${$r_attachment}{encoding} =~ /^uuencode$/i) {
#            $content = uudecode(${$r_attachment}{contents});
         } else { ## Guessing it's 7-bit, at least sending SOMETHING back! :)
            $content = ${$r_attachment}{contents};
         }

         if (${$r_attachment}{contenttype} =~ m#^text/html#i ) {
            my $escapedmessageid = &urlencode($messageid);
            $content = html4nobase($content);
            $content = html4attachments($content, $r_attachments, $cgi_url, "step=mail&amp;action=viewattachment&amp;message_id=$escapedmessageid&amp;folder=$escapedfolder");
            $content = html4mailto($content, $cgi_url, "step=mail&amp;action=composemessage&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;composetype=sendto");
         }

         my $length = length($content);
         print qq|Content-Length: $length\n|.
               qq|Content-Transfer-Coding: binary\n|.
               qq|Connection: close\n|.
               qq|Content-Type: ${$r_attachment}{contenttype}; name="${$r_attachment}{filename}"\n|;
         unless (${$r_attachment}{contenttype} =~ /^text/i) {
            print qq|Content-Disposition: attachment; filename="${$r_attachment}{filename}"\n|;
         }
         print qq|\n|.
               qq|$content|;
      } else {
         printheader();
         print "What the heck? Message seems to be gone!";
         printfooter();
      }
      return;

   }
}

################### END VIEWATTACHMENT ##################

################## GETHEADERS #######################
sub getheaders {
   my @message = ();
   my $spoolfile;
   my $headerdb;
   my $spoolhandle=FileHandle->new();
   my @messageids;
   my $messageid;

   ($spoolfile, $headerdb)=get_spoolfile_headerdb_and_set_uid_gid();

   unless ( filelock($spoolfile, LOCK_SH|LOCK_NB) ) {
      neomailerror("$lang_err{'couldnt_locksh'} $spoolfile!");
   }

   update_headerdb($headerdb, $spoolfile);

   if ( $keyword ne '' ) {
      open($spoolhandle, "$spoolfile");
      my %haskeyword=search_messages_for_keyword($keyword, $headerdb, $spoolhandle, "$folderdir/.search.cache");
      close($spoolhandle);
      foreach ( get_messageids_sorted($headerdb, $sort, "$headerdb.cache") ) {
         push (@messageids, $_) if ( $haskeyword{$_} == 1 ); 
      }
   } else {
      @messageids=get_messageids_sorted($headerdb, $sort, "$headerdb.cache");
   }

   filelock($spoolfile, LOCK_UN);
   chown($menu_id,$gid,$spoolfile);

   $messagenumber=-1;
   $total_size=0;

   filelock("$headerdb.$dbm_ext", LOCK_SH);
   dbmopen (%HDB, $headerdb, undef);

   foreach $messageid (@messageids) {
### Get a unique memory address before creating pointer
      my %header;
            
      $header{'message_id'}=$messageid;
      ( $header{'offset'},
	$header{'from'}, $header{'to'}, $header{'date'}, $header{'subject'},
	$header{'content_type'}, $header{'status'}, 
	$header{'messagesize'} )=split(/@@@/, $HDB{$messageid});

      # don't show the imap control message in INBOX to user
      if ( ($folder eq "INBOX") && ($header{'subject'} =~ /DON'T DELETE THIS MESSAGE/) ) {
         undef(%header);
         next;
      }

      $total_size+=$header{'messagesize'};
      $messagenumber++;
      $message[$messagenumber] = \%header;
   }

   dbmclose(%HDB);
   filelock("$headerdb.$dbm_ext", LOCK_UN);

   return \@message;
}

#################### END GETHEADERS #######################

#################### GETMESSAGE ###########################
sub getmessage {
   my $messageid = $_[0];
   my $spoolfile;
   my $headerdb;
   my $spoolhandle=FileHandle->new();
   my @messageids;

   my %message = ();

   my ($currentheader, $currentbody, $r_currentattachments, $currentfrom, $currentdate,
       $currentsubject, $currentid, $currenttype, $currentto, $currentcc,
       $currentreplyto, $currentencoding, $currentstatus);

   ($spoolfile, $headerdb)=get_spoolfile_headerdb_and_set_uid_gid();

   unless(filelock($spoolfile, LOCK_SH|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_locksh'} $spoolfile!");
   }
   update_headerdb($headerdb, $spoolfile);
   open($spoolhandle, "$spoolfile");

   if ( $keyword ne '' ) {
      my %haskeyword=search_messages_for_keyword($keyword, $headerdb, $spoolhandle, "$folderdir/.search.cache");
      foreach ( get_messageids_sorted($headerdb, $sort, "$headerdb.cache") ) {
         push (@messageids, $_) if ( $haskeyword{$_} == 1 ); 
      }
   } else {
      @messageids=get_messageids_sorted($headerdb, $sort, "$headerdb.cache");
   }

   # $r_attachment is a reference to attachment array!
   ($currentheader, $currentbody, $r_currentattachments)
	=parse_rfc822block(get_message_block($messageid, $headerdb, $spoolhandle));

   close($spoolhandle);
   filelock($spoolfile, LOCK_UN);
   chown($menu_id,$gid,$spoolfile);

   return \%message if ( $currentheader eq "" );

   $currentfrom = $currentdate = $currentsubject = $currenttype = 
   $currentto = $currentcc = $currentreplyto = $currentencoding = 'N/A';
   $currentstatus = '';

   my $lastline = 'NONE';
   foreach (split(/\n/, $currentheader)) {
      if (/^\s/) {
         if    ($lastline eq 'FROM') { $currentfrom .= $_ }
         elsif ($lastline eq 'REPLYTO') { $currentreplyto .= $_ }
         elsif ($lastline eq 'DATE') { $currentdate .= $_ }
         elsif ($lastline eq 'SUBJ') { $currentsubject .= $_ }
         elsif ($lastline eq 'MESSID') { s/^\s+//; $currentid .= $_ }
         elsif ($lastline eq 'TYPE') { $currenttype .= $_ }
         elsif ($lastline eq 'ENCODING') { $currentencoding .= $_ }
         elsif ($lastline eq 'TO')   { $currentto .= $_ }
         elsif ($lastline eq 'CC')   { $currentcc .= $_ }
      } elsif (/^from:\s+(.+)$/ig) {
         $currentfrom = $1;
         $lastline = 'FROM';
      } elsif (/^reply-to:\s+(.+)$/ig) {
         $currentreplyto = $1;
         $lastline = 'REPLYTO';
      } elsif (/^to:\s+(.+)$/ig) {
         $currentto = $1;
         $lastline = 'TO';
      } elsif (/^cc:\s+(.+)$/ig) {
         $currentcc = $1;
         $lastline = 'CC';
      } elsif (/^date:\s+(.+)$/ig) {
         $currentdate = $1;
         $lastline = 'DATE';
      } elsif (/^subject:\s+(.+)$/ig) {
         $currentsubject = $1;
         $lastline = 'SUBJ';
      } elsif (/^message-id:\s+(.*)$/ig) {
         $currentid = $1;
         $lastline = 'MESSID';
      } elsif (/^content-type:\s+(.+)$/ig) {
         $currenttype = $1;
         $lastline = 'TYPE';
      } elsif (/^content-transfer-encoding:\s+(.+)$/ig) {
         $currentencoding = $1;
         $lastline = 'ENCODING';
      } elsif (/^status:\s+(.+)$/ig) {
         $currentstatus = $1;
         $lastline = 'NONE';
      } else {
         $lastline = 'NONE';
      }
   }

   $message{header} = $currentheader;
   $message{body} = $currentbody;
   $message{attachment} = $r_currentattachments;

   $message{from}    = decode_mimewords($currentfrom);
   $message{replyto} = decode_mimewords($currentreplyto) unless ($currentreplyto eq "N/A");
   $message{to}      = decode_mimewords($currentto) unless ($currentto eq "N/A");
   $message{cc}      = decode_mimewords($currentcc) unless ($currentcc eq "N/A");
   $message{subject} = decode_mimewords($currentsubject);

   $message{date} = $currentdate;
   $message{status} = $currentstatus;
   $message{messageid} = $currentid;
   $message{contenttype} = $currenttype;
   $message{encoding} = $currentencoding;

   # Determine message's number and previous and next message IDs.
   foreach my $messagenumber (0..$#messageids) {
      if ($messageids[$messagenumber] eq $messageid) {
         $message{"prev"} = $messageids[$messagenumber-1] if ($messagenumber > 0);
         $message{"next"} = $messageids[$messagenumber+1] if ($messagenumber < $#messageids);
         $message{"number"} = $messagenumber+1;
         $message{"total"}=$#messageids+1;
         last;
      }
   }
   return \%message;
}

#################### END GETMESSAGE #######################

#################### UPDATESTATUS #########################
sub updatestatus {
   my ($messageid, $status) = @_;
   my ($currmessage, $currentheader, $currentbody);
   my $spoolfile;
   my $headerdb;
   my $spoolhandle=FileHandle->new();

   ($spoolfile, $headerdb)=get_spoolfile_headerdb_and_set_uid_gid();

# since spool must exists here, we do lock before open
   unless (filelock($spoolfile, LOCK_EX|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_lock'} $spoolfile!");
   }
   update_headerdb($headerdb, $spoolfile);
   open ($spoolhandle, "+<$spoolfile") or 
	neomailerror("$lang_err{'couldnt_open'} $spoolfile!");

   my @messageids=get_messageids_sorted_by_offset($headerdb);
   my $movement;
   my @attr;
   my $i;

   filelock("$headerdb.$dbm_ext", LOCK_EX);
   dbmopen (%HDB, $headerdb, 600);

   for ($i=0; $i<=$#messageids; $i++) {
      if ($messageids[$i] eq $messageid) {
         @attr=split(/@@@/, $HDB{$messageid});

         my $messagestart=$attr[$_OFFSET];
         my $messagesize=$attr[$_SIZE];
         my $messageend=$messagestart+$messagesize;
         my $foldersize;
         my ($messagenewsize, $messagenewstatus);

         seek ($spoolhandle, $messagestart, 0) or neomailerror("$lang_err{'couldnt_seek'} $spoolfile!");
         read($spoolhandle, $currmessage, $messagesize);

         ($currentheader, $currentbody) = split(/\n\r*\n/,$currmessage, 2);

         if ($currentheader =~ s/^status:\s?(.*?)$/Status: $status$1/im) {
           $messagenewstatus="$status$1";
         } else {
           $currentheader .= "\nStatus: $status";
           $messagenewstatus="$status";
         }
         $currentheader="From $currentheader" unless ($currentheader =~ /^From /);

         $messagenewsize=length($currentheader)+length("\n\n")+length($currentbody);
         $movement=$messagenewsize-$messagesize;

         my @s=stat($spoolhandle); $foldersize=$s[7];
         shiftblock($spoolhandle, $messageend, $foldersize-$messageend, $movement);

         seek($spoolhandle, $messagestart, 0) or neomailerror("$lang_err{'couldnt_seek'} $spoolfile!");
         print $spoolhandle $currentheader, "\n\n", $currentbody;

         seek($spoolhandle, $foldersize+$movement, 0);
         truncate($spoolhandle, tell($spoolhandle));

### set attributes in headerdb for this status changed message
         if ($attr[$_STATUS]!~/r/i && $messagenewstatus=~/r/i) {
            $HDB{'NEWMESSAGES'}--;
         }
         $attr[$_SIZE]=$messagenewsize;
         $attr[$_STATUS]=$messagenewstatus;
         $HDB{$messageid}=join('@@@', @attr);

         last;
      }
   }
   $i++;

### set attributes in headerdb for messages after the above status changed message
   for (;$i<=$#messageids; $i++) {
      @attr=split(/@@@/, $HDB{$messageids[$i]});
      $attr[$_OFFSET]+=$movement;
      $HDB{$messageids[$i]}=join('@@@', @attr);
   }

   close ($spoolhandle) or neomailerror("$lang_err{'couldnt_close'} $spoolfile!");

   $HDB{'METAINFO'}=metainfo($spoolfile);
   dbmclose(%HDB);
   filelock("$headerdb.$dbm_ext", LOCK_UN);

   filelock("$spoolfile", LOCK_UN);
   chown($menu_id,$gid,$spoolfile);
}
################### END UPDATESTATUS ######################

#################### MOVEMESSAGE ########################
sub movemessage {
   my $destination = $DATA{"destination"};
   ($destination =~ /(.+)/) && ($destination = $1);
   my @messageids = split(/,/,$DATA{"message_ids"});
   my $messageids = join("\n", @messageids);
   my $spoolfile;
   my $headerdb;
   my $spoolhandle=FileHandle->new();

   if ( $#messageids<0 ) {	# no message ids to delete, return immediately
      if ($DATA{"messageaftermove"}) {
         readmessage();
      } else {
         displayheaders();
      }
      return;
   }

   neomailerror ("$lang_err{'shouldnt_move_here'}") if 
      (($folder eq $destination) || ($destination eq 'INBOX'));

   unless ( ($destination eq 'TRASH') || ($destination eq 'SAVED') ||
        ($destination eq 'SENT') || ($destination eq 'sent_mail') ||
        ($destination eq 'saved_messages') ||
        ($destination eq 'neomail_trash') ) {
      $destination .= ".folder" unless ( $homedirfolders eq 'yes' );
      neomailerror("$lang_err{'destination_folder'} $destination $lang_err{'doesnt_exist'}") unless 
         ( -f "$folderdir/$destination" );
   }

   if ($hitquota) {
      unless ( ($destination eq 'TRASH') || ($destination eq 'neomail_trash') ){
         neomailerror("$lang_err{'folder_hitquota'}");
      }
   }

   ($spoolfile, $headerdb)=get_spoolfile_headerdb_and_set_uid_gid();

# open source folder, since spool must exist => lock before open
   unless (filelock($spoolfile, LOCK_EX|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_lock'} $spoolfile!");
   }
   update_headerdb($headerdb, $spoolfile);
   open ($spoolhandle, "+<$spoolfile") or 
	neomailerror("$lang_err{'couldnt_open'} $spoolfile!");

# open destination folder, since dest may not exist => open before lock
   open (DEST, ">>$folderdir/$destination") or
      neomailerror ("$lang_err{'couldnt_open'} $destination!");
   unless (filelock("$folderdir/$destination", LOCK_EX|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_lock'} $destination!");
   }
   update_headerdb("$folderdir/.$destination", "$folderdir/$destination");

   my @allmessageids=get_messageids_sorted_by_offset($headerdb);
   my ($blockstart, $blockend, $writepointer);
   my ($currmessage, $messagestart, $messagesize, @a);
   my $moved=0;
   
   filelock("$headerdb.$dbm_ext", LOCK_EX);
   dbmopen (%HDB, $headerdb, 600);
   filelock("$folderdir/.$destination.$dbm_ext", LOCK_EX);
   dbmopen (%HDB2, "$folderdir/.$destination", 600);

   $blockstart=$blockend=$writepointer=0;

   for (my $i=0; $i<=$#allmessageids; $i++) {
      @attr=split(/@@@/, $HDB{$allmessageids[$i]});

      if ($messageids =~ /^\Q$allmessageids[$i]\E$/m) {	# msg to be moved
         $moved++;

         $messagestart=$attr[$_OFFSET];
         $messagesize=$attr[$_SIZE];

         shiftblock($spoolhandle, $blockstart, $blockend-$blockstart, $writepointer-$blockstart);

         $writepointer=$writepointer+($blockend-$blockstart);
         $blockstart=$blockend=$messagestart+$messagesize;

         seek($spoolhandle, $attr[$_OFFSET], 0);
         read($spoolhandle, $currmessage, $attr[$_SIZE]);
         
         unless ($hitquota) {	# messages to neo_trash are DROPED directly if hitquota!!!
            $attr[$_OFFSET]=tell(DEST);
            if ($currmessage =~ /^From /) {
               $attr[$_SIZE]=length($currmessage);
               print DEST $currmessage;
            } else {
               $attr[$_SIZE]=length("From ")+length($currmessage);
               print DEST "From ", $currmessage;
            }
            if ( $attr[$_STATUS]!~/r/i ) {
               $HDB2{'NEWMESSAGES'}++;
            }
            $HDB2{'ALLMESSAGES'}++;
            $HDB2{$allmessageids[$i]}=join('@@@', @attr);
         } 
         
         if ( $attr[$_STATUS]!~/r/i ) {
            $HDB{'NEWMESSAGES'}--;
         }
         $HDB{'ALLMESSAGES'}--;
         delete $HDB{$allmessageids[$i]};

      } else {						# msg to be kept in same folder
         $messagestart=$attr[$_OFFSET];
         $messagesize=$attr[$_SIZE];
         $blockend=$messagestart+$messagesize;

         my $movement=$writepointer-$blockstart;
         if ($movement<0) {
            $attr[$_OFFSET]+=$movement;
            $HDB{$allmessageids[$i]}=join('@@@', @attr);
         }
      }
   }

   if ($moved>0) {
      shiftblock($spoolhandle, $blockstart, $blockend-$blockstart, $writepointer-$blockstart);
      seek($spoolhandle, $writepointer+($blockend-$blockstart), 0);
      truncate($spoolhandle, tell($spoolhandle));
   }

   close (DEST) or neomailerror("$lang_err{'couldnt_close'} $destination!");
   close ($spoolhandle) or neomailerror("$lang_err{'couldnt_close'} $spoolfile!");

   $HDB2{'METAINFO'}=metainfo("$folderdir/$destination");
   dbmclose(%HDB2);
   filelock("$folderdir/.$destination.$dbm_ext", LOCK_UN);

   $HDB{'METAINFO'}=metainfo($spoolfile);
   dbmclose(%HDB);
   filelock("$headerdb.$dbm_ext", LOCK_UN);

   filelock("$folderdir/$destination", LOCK_UN);
   filelock($spoolfile, LOCK_UN);
   chown($menu_id,$gid,$spoolfile);

   if ($moved>0) {
      $messageids =~ s/\n/, /g;
      writelog("moved messages to $destination - ids=$messageids");
   }

   if ($DATA{"messageaftermove"}) {
      readmessage();
   } else {
      displayheaders();
   }
}
#################### END MOVEMESSAGE #######################

#################### EMPTYTRASH ########################
sub emptytrash {
   my $trashfile;
   if ( $homedirfolders eq 'yes' ) {
      $trashfile = "$folderdir/neomail_trash";
   } else {
      $trashfile = "$folderdir/TRASH";
   }
#   if ( ($homedirfolders eq 'yes') && ($> == 0) ) {
#      $) = $gid;
#      $> = $menu_id;
#      umask(0077); # make sure only owner can read/write
#   }
   open (TRASH, ">$trashfile") or
      neomailerror ("$lang_err{'couldnt_open'} $trashfile!");
   close (TRASH) or neomailerror("$lang_err{'couldnt_close'} $trashfile!");
   writelog("trash emptied");
   displayheaders();
}
#################### END EMPTYTRASH #######################


##################### GETATTLIST ###############################
sub getattlist {
   my $currentfile;
   my @attlist;
   $savedattsize = 0;
   opendir (NEOMAILDIR, "$neomaildir") or
      neomailerror("$lang_err{'couldnt_open'} $neomaildir!");
   while (defined($currentfile = readdir(NEOMAILDIR))) {
      if ($currentfile =~ /^($UIDNM{$menu_id}-att\d+)$/) {
         $currentfile = $1;
         $savedattsize += ( -s "$neomaildir$currentfile" );
         open (ATTFILE, "$neomaildir$currentfile");
         while (defined(my $line = <ATTFILE>)) {
            if ($line =~ s/^.+name="?([^"]+)"?.*$/$1/i) {
               $line = str2html($line);
               $line =~ s/^(.*)$/<em>$1<\/em>/;
               push (@attlist, $line);
               last;
            }
         }
         close (ATTFILE);
      }
   }
   closedir (NEOMAILDIR);
   return \@attlist;
}
##################### END GETATTLIST ###########################

##################### DELETEATTACHMENTS ############################
sub deleteattachments {
   my $currentfile;
   opendir (NEOMAILDIR, "$neomaildir") or
      neomailerror("$lang_err{'couldnt_open'} $neomaildir!");
   while (defined($currentfile = readdir(NEOMAILDIR))) {
      if ($currentfile =~ /^($UIDNM{$menu_id}-att\d+)$/) {
         $currentfile = $1;
         unlink ("$neomaildir$currentfile");
      }
   }
   closedir (NEOMAILDIR);
}
#################### END DELETEATTACHMENTS #########################

##################### FIRSTTIMEUSER ################################
sub firsttimeuser {
   my $html = '';
   my $temphtml;
   open (FTUSER, "$neomaildir/templates/$lang/firsttimeuser.template") ||
      neomailerror("$lang_err{'couldnt_open'} firsttimeuser.template!");
   while (<FTUSER>) {
      $html .= $_;
   }
   close (FTUSER);
   
   $html = applystyle($html);

   printheader();

   $temphtml = '<form method=POST action="'.$cgi_url.'">';
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='action' value='setprefs'>";
   $temphtml .= "<input type='hidden' name='firsttimeuser' value='yes'>";
   $temphtml .= "<input type='hidden' name='realname' value='".$UIDNM{$menu_id}."'>";
   $temphtml .= "<input type='submit' name='".$lang_text{'continue'}."' value='".$lang_text{'continue'}."'>";
   $temphtml .= "</form>";
   $html =~ s/\@\@\@CONTINUEBUTTON\@\@\@/$temphtml/;
   
   print $html;

   printfooter();
}
################### END FIRSTTIMEUSER ##############################

################## RETRIVE POP3 ###########################
sub retrpop3 {
   my ($spoolfile, $pop3host, $pop3user);
   my (%account, $response);

   # get $spoolfile from globals
   if ($homedirspools eq "yes") {
      $spoolfile = "$homedir/$homedirspoolname";
   } elsif ($hashedmailspools eq "yes") {
      $user = $UIDNM{$menu_id};
      $user =~ /^(.)(.)/;
      my $firstchar = $1;
      my $secondchar = $2;
      $spoolfile = "$mailspooldir/$firstchar/$secondchar/$user";
   } else {
      $spoolfile = "$mailspooldir/$UIDNM{$menu_id}";
   }

   # create system spool file /var/mail/xxxx
   if ( ! -f "$spoolfile" ) {
      open (F, ">>$spoolfile");
      close(F);
      chown($menu_id,$gid,$spoolfile);
      chmod(0600, $spoolfile);
   }

   $pop3host = $DATA{"host"} || '';
   $pop3user = $DATA{"name"} || '';

   if ( ! -f "$folderdir/.pop3.book" ) {
      print "Location:  $scripturl&amp;action=editpop3&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage";
   }

   %account = getpop3book("$folderdir/.pop3.book");
   if (%account == 0) {
      neomailerror("$lang_err{'couldnt_open'} $folderdir/.pop3.book!");
   }

#   unless (filelock($spoolfile, LOCK_EX|LOCK_NB)) {
#      neomailerror("$lang_err{'couldnt_lock'} $spoolfile!");
#   }
   $response = retrpop3mail($pop3host, $pop3user, 
				"$folderdir/.pop3.book", $spoolfile);
#   filelock($spoolfile, LOCK_UN);

   if ($response>=0) {	# new mail found
      $folder="INBOX";
      print "Location: $scripturl&action=displayheaders&sort=$sort&firstmessage=$firstmessage&folder=$folder\n\n";
   } else {
      writelog("pop3 error $response at $pop3user\@$pop3host");
      if ($response == -1) {
   	  neomailerror("$lang_err{'user_not_exist'}");
      } elsif ($response == -2) {
   	  neomailerror("$lang_err{'couldnt_open'} pop3 folder for $host");
      } elsif ($response == -3) {
   	  neomailerror("$host $lang_err{'network_server_error'}");
      } elsif ($response == -4) {
      	  neomailerror("$host $lang_err{'password_error'}");
      }
   }
}
################## END RETRIVE POP3 ###########################

################## RETRIVE ALL POP3 ###########################
sub retrpop3s {
   my (%account, $response);
   my $spoolfile;
   my $fetch_complete=0;
   my $i;

   # get $spoolfile from globals
   if ($homedirspools eq "yes") {
      $spoolfile = "$homedir/$homedirspoolname";
   } elsif ($hashedmailspools eq "yes") {
      $user = $UIDNM{$menu_id};
      $user =~ /^(.)(.)/;
      my $firstchar = $1;
      my $secondchar = $2;
      $spoolfile = "$mailspooldir/$firstchar/$secondchar/$user";
   } else {
      $spoolfile = "$mailspooldir/$UIDNM{$menu_id}";
   }

   # create system spool file /var/mail/xxxx
   if ( ! -f "$spoolfile" ) {
      open (F, ">>$spoolfile");
      close(F);
      chown($menu_id,$gid,$spoolfile);
      chmod(0600, $spoolfile);
   }

   if ( ! -f "$folderdir/.pop3.book" ) {
      print "Location: $scripturl&amp;action=editpop3&amp;sort=$sort&amp;keyword=$escapedkeyword&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage";
   }

   %account = getpop3book("$folderdir/.pop3.book");

   # fork a child to do fetch pop3 mails and return immediately
   if (%account >0) {
      $|=1; 				# flush all output
      $SIG{CHLD} = sub { wait; $fetch_complete=1; };	# handle zombie

      if ( fork() == 0 ) {		# child
         close(STDOUT);
         close(STDIN);

         foreach (values %account) {
            my ($pop3host, $pop3user, $mbox);
            my ($response, $dummy);

            ($pop3host, $pop3user, $dummy) = split(/:/,$_, 3);
            $response = retrpop3mail($pop3host, $pop3user, 
         				"$folderdir/.pop3.book",  $spoolfile);
            if ( $response<0) {
               writelog("pop3 error $response at $pop3user\@$pop3host");
            }
         }
         exit;
      }
   }

   for ($i=0; $i<10; $i++) {	# wait fetch to complete for 10 seconds
      sleep 1;
      if ($fetch_complete==1) {
         last;
      }
   }   

   $folder="INBOX";
   print "Location: $scripturl&action=displayheaders&sort=$sort&firstmessage=$firstmessage&folder=$folder\n\n";
}
################## END RETRIVE ALL POP3 ###########################

###################### READPREFS #########################
sub readprefs {
   my ($key,$value);
   my %prefshash;
   if ( -f "$userprefsdir$UIDNM{$menu_id}/config" ) {
      open (CONFIG,"$userprefsdir$UIDNM{$menu_id}/config") or
         neomailerror("$lang_err{'couldnt_open'} config!");
      while (<CONFIG>) {
         ($key, $value) = split(/=/, $_);
         chomp($value);
         if ($key eq 'style') {
            $value =~ s/\.//g;  ## In case someone gets a bright idea...
         }
         $prefshash{"$key"} = $value;
      }
      close (CONFIG) or neomailerror("$lang_err{'couldnt_close'} config!");
   }
   if ( -f "$userprefsdir$UIDNM{$menu_id}/signature" ) {
      $prefshash{"signature"} = '';
      open (SIGNATURE, "$userprefsdir$UIDNM{$menu_id}/signature") or
         neomailerror("$lang_err{'couldnt_open'} signature!");
      while (<SIGNATURE>) {
         $prefshash{"signature"} .= $_;
      }
      close (SIGNATURE) or neomailerror("$lang_err{'couldnt_close'} signature!");
   }
   return \%prefshash;
}
##################### END READPREFS ######################

###################### READSTYLE #########################
sub readstyle {
   my ($key,$value);
   my $stylefile = $prefs{"style"} || 'Default';
   my %stylehash;
   unless ( -f "$stylesdir$stylefile") {
      $stylefile = 'Default';
   }
   open (STYLE,"$stylesdir$stylefile") or
      neomailerror("$lang_err{'couldnt_open'} $stylefile!");
   while (<STYLE>) {
      if (/###STARTSTYLESHEET###/) {
         $stylehash{"css"} = '';
         while (<STYLE>) {
            $stylehash{"css"} .= $_;
         }
      } else {
         ($key, $value) = split(/=/, $_);
         chomp($value);
         $stylehash{"$key"} = $value;
      }
   }
   close (STYLE) or neomailerror("$lang_err{'couldnt_close'} $stylefile!");
   return \%stylehash;
}
##################### END READSTYLE ######################

##################### WRITELOG ############################
sub writelog {
   unless ( ($logfile eq 'no') || ( -l "$logfile" ) ) {
	  $> = 0;
      open (LOGFILE,">>$logfile") or neomailerror("$lang_err{'couldnt_open'} $logfile!");
      my $timestamp = localtime();
      my $logaction = shift;
      my $loggeduser = $UIDNM{$menu_id} || 'UNKNOWNUSER';
      print LOGFILE "$timestamp - [$$] ($userip) $loggeduser - $logaction\n";
      close (LOGFILE);
   }
}
#################### END WRITELOG #########################

##################### NEOMAILERROR ##########################
sub neomailerror {
   unless ($headerprinted) {
      $headerprinted = 1;
      my $background = $style{"background"};
      $background =~ s/"//g;

      print "Content-type: text/html\n\n";
      print "<html><head>\n";
      print "<TITLE>NeoMail Professional $version</TITLE>\n";
      print '<META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset='.$lang_charset.'">'."\n";
      print '<META HTTP-EQUIV="Pargma" CONTENT="no-cache">'."\n";
      print "</HEAD><BODY BACKGROUND=\"$bg_url\" BGCOLOR=$background>\n";
      print '<style type="text/css">';
      print $style{"css"};
      print '</style>';
      print "<FONT FACE=",$style{"fontface"},">\n";
   }
   print '<BR><BR><BR><BR><BR><BR>';
   print '<table border="0" align="center" width="40%" cellpadding="1" cellspacing="1">';

   print '<tr><td bgcolor=',$style{"titlebar"},' align="left">',
   '<font color=',$style{"titlebar_text"},' face=',$style{"fontface"},' size="3"><b>NEOMAIL ERROR</b></font>',
   '</td></tr><tr><td align="center" bgcolor=',$style{"window_light"},'><BR>',$>,':';
   print shift;
   print '<BR><BR></td></tr></table></BODY></HTML>';
   exit 0;
}
################### END NEOMAILERROR #######################

##################### PRINTHEADER #########################
sub printheader {
   unless ($headerprinted) {
      my $html = '';
      $headerprinted = 1;
      open (HEADER, "$neomaildir/templates/$lang/header.template") or
         neomailerror("$lang_err{'couldnt_open'} header.template!");
      while (<HEADER>) {
         $html .= $_;
      }
      close (HEADER);
	  $mytitle = "<center><font size=+2 face=з color=darkblue>l $myver</font> [ ";
	  if ($admin eq '1') {
		$mytitle .= "<a href=/help/help_root.htm target=_blank>޲zϥλ</a> ]</center>";
	  } else {
		$mytitle .= "<a href=/help/help_user.htm target=_blank>ϥλ</a> ]</center>";
	  }
      $html = applystyle($html);
	  $html =~ s/\@\@\@BG_URL\@\@\@/$bg_url/g;
      $html =~ s/\@\@\@CHARSET\@\@\@/$lang_charset/g;
      $html =~ s/\@\@\@TITLE\@\@\@/$mytitle/g;
      print "Content-type: text/html\n\n";
      print $html;
   }
}
################### END PRINTHEADER #######################

################### PRINTFOOTER ###########################
sub printfooter {
   my $html = '';
   open (FOOTER, "$neomaildir/templates/$lang/footer.template") or
      neomailerror("$lang_err{'couldnt_open'} footer.template!");
   while (<FOOTER>) {
      $html .= $_;
   }
   close (FOOTER);
   
   $html = applystyle($html);
   
   $html =~ s/\@\@\@VERSION\@\@\@/$version/g;
   print $html;
   &foot;
}
################# END PRINTFOOTER #########################

################# APPLYSTYLE ##############################
sub applystyle {
   my $template = shift;
   $template =~ s/\@\@\@LOGO_URL\@\@\@/$logo_url/g;
   $template =~ s/\@\@\@BACKGROUND\@\@\@/$style{"background"}/g;
   $template =~ s/\@\@\@TITLEBAR\@\@\@/$style{"titlebar"}/g;
   $template =~ s/\@\@\@TITLEBAR_TEXT\@\@\@/$style{"titlebar_text"}/g;
   $template =~ s/\@\@\@MENUBAR\@\@\@/$style{"menubar"}/g;
   $template =~ s/\@\@\@WINDOW_DARK\@\@\@/$style{"window_dark"}/g;
   $template =~ s/\@\@\@WINDOW_LIGHT\@\@\@/$style{"window_light"}/g;
   $template =~ s/\@\@\@ATTACHMENT_DARK\@\@\@/$style{"attachment_dark"}/g;
   $template =~ s/\@\@\@ATTACHMENT_LIGHT\@\@\@/$style{"attachment_light"}/g;
   $template =~ s/\@\@\@COLUMNHEADER\@\@\@/$style{"columnheader"}/g;
   $template =~ s/\@\@\@TABLEROW_LIGHT\@\@\@/$style{"tablerow_light"}/g;
   $template =~ s/\@\@\@TABLEROW_DARK\@\@\@/$style{"tablerow_dark"}/g;
   $template =~ s/\@\@\@FONTFACE\@\@\@/$style{"fontface"}/g;
   $template =~ s/\@\@\@SCRIPTURL\@\@\@/$scripturl/g;
   $template =~ s/\@\@\@scripturl\@\@\@/$scripturl/g;
   $template =~ s/\@\@\@CSS\@\@\@/$style{"css"}/g;
   $template =~ s/\@\@\@VERSION\@\@\@/$version/g;

   return $template;
}
################ END APPLYSTYLE ###########################

################# FILTERMESSAGE ###########################
sub filtermessage {
   ## variables ##
   my @filterrules;
   my ($spoolfile, $headerdb);
   my $spoolhandle=FileHandle->new();
   my (%HDB, %HDB2);
   my ($blockstart, $blockend, $writepointer);
   my (@allmessageids, $i);
   my $moved=0;
   
   ## get $spoolfile ##
   if ($homedirspools eq "yes") {
      $spoolfile = "$homedir/$homedirspoolname";
   } elsif ($hashedmailspools eq "yes") {
      $user = $UIDNM{$menu_id};
      $user =~ /^(.)(.)/;
      my $firstchar = $1;
      my $secondchar = $2;
      $spoolfile = "$mailspooldir/$firstchar/$secondchar/$user";
   } else {
      $spoolfile = "$mailspooldir/$UIDNM{$menu_id}";
   }
   
   ## check existence of spoolfile
   if ( ! -f $spoolfile ) {
      return;
   }
   ## check .filter_check ##
   if ( -f "$folderdir/.filter.check" ) {
      open (FILTERCHECK, "$folderdir/.filter.check" ) or
            neomailerror("$lang_err{'couldnt_open'} .filter.check!");
      if (<FILTERCHECK> eq metainfo($spoolfile)) {
         return;
      }      
      close (FILTERCHECK) or neomailerror("$lang_err{'couldnt_close'} .filter.check!");
   }

   ## get $headerdb ##
   $headerdb="$folderdir/.$UIDNM{$menu_id}";
   
   ## get @filterrules ##
   if ( -f "$folderdir/.filter.book" ) {
      open (FILTER,"$folderdir/.filter.book") or
         neomailerror("$lang_err{'couldnt_open'} .filter.book!");
      while (<FILTER>) {
         chomp($_);
         push (@filterrules, $_);
      }
      close (FILTER) or neomailerror("$lang_err{'couldnt_close'} .filter.book!");
   }

   ## open INBOX, since spool must exist => lock before open ##
   unless (filelock($spoolfile, LOCK_EX|LOCK_NB)) {
      neomailerror("$lang_err{'couldnt_lock'} $spoolfile!");
   }
   update_headerdb($headerdb, $spoolfile);
   open ($spoolhandle, "+<$spoolfile") or neomailerror("$lang_err{'couldnt_open'} $spoolfile!");

   @allmessageids=get_messageids_sorted_by_offset($headerdb);

   ## open INBOX dbm => lock before open ##
   filelock("$headerdb.$dbm_ext", LOCK_EX);
   dbmopen (%HDB, $headerdb, 600);

   $blockstart=$blockend=$writepointer=0;

   for ($i=0; $i<=$#allmessageids; $i++) {
      my ($priority, $rules, $include, $text, $destination, $enable);
      my ($messagestart, $messagesize);
      my @attr = split(/@@@/, $HDB{$allmessageids[$i]});
      my ($currmessage, $header, $body, $r_attachments)=("", "", "", "");
      my $is_message_moved = 0;
      
      ## if match filterrules => move message ##
      foreach (sort @filterrules) {
         ($priority, $rules, $include, $text, $destination, $enable) = split(/\@\@\@/, $_);
         
         ## check is currentrule enable ##
         next unless ($enable == 1);
         
         if ( $rules eq 'from' ) {
            if (   ($include eq 'include' && $attr[$_FROM] =~ /$text/i)
                || ($include eq 'exclude' && $attr[$_FROM] !~ /$text/i)  ) {
               $is_message_moved = 1;
               last;
            }

         } elsif ( $rules eq 'to' ) {
            if (   ($include eq 'include' && $attr[$_TO] =~ /$text/i)
                || ($include eq 'exclude' && $attr[$_TO] !~ /$text/i)  ) {
               $is_message_moved = 1;
               last;
            }

         } elsif ( $rules eq 'subject' ) {
            if (   ($include eq 'include' && $attr[$_SUBJECT] =~ /$text/i)
                || ($include eq 'exclude' && $attr[$_SUBJECT] !~ /$text/i)  ) {
               $is_message_moved = 1;
               last;
            }

         } elsif ( $rules eq 'header' ) {
            if ($currmessage eq "") {
               seek($spoolhandle, $attr[$_OFFSET], 0);
               read($spoolhandle, $currmessage, $attr[$_SIZE]);
            }
            if ($header eq "") {
               ($header, $body, $r_attachments)=parse_rfc822block(\$currmessage);
            }

            $header=decode_mimewords($header);
            if (  ( $include eq 'include' && $header =~ /$text/i )
                ||( $include eq 'exclude' && $header !~ /$text/i ) ) {
               $is_message_moved = 1;
               last;
            }

         } elsif ( $rules eq 'body' ) {
            if ($currmessage eq "") {
               seek($spoolhandle, $attr[$_OFFSET], 0);
               read($spoolhandle, $currmessage, $attr[$_SIZE]);
            }
            if ($header eq "") {
               ($header, $body, $r_attachments)=parse_rfc822block(\$currmessage);
            }

            if ( $attr[$_CONTENT_TYPE] =~ /^text/i ) {	# read all for text/plain. text/html
               if ( $header =~ /content-transfer-encoding:\s+quoted-printable/i) {
                  $body = decode_qp($body);
               } elsif ($header =~ /content-transfer-encoding:\s+base64/i) {
                  $body = decode_base64($body);
               }
            }
            if (  ( $include eq 'include' && $body =~ /$text/i )
                ||( $include eq 'exclude' && $body !~ /$text/i ) ) { # o for regex, ms for ^ $ match with line
               $is_message_moved = 1;
               last;
            }

            # check attachments body
            foreach my $r_attachment (@{$r_attachments}) {
               if ( ${$r_attachment}{contenttype} =~ /^text/i ) {	# read all for text/plain. text/html
                  if ( ${$r_attachment}{encoding} =~ /^quoted-printable/i ) {
                     ${$r_attachment}{contents} = decode_qp( ${$r_attachment}{contents});
                  } elsif ( ${$r_attachment}{encoding} =~ /^base64/i ) {
                     ${$r_attachment}{contents} = decode_base64( ${$r_attachment}{contents});
                  }
                  if (  ( $include eq 'include' && ${$r_attachment}{contents} =~ /$text/i )
                      ||( $include eq 'exclude' && ${$r_attachment}{contents} !~ /$text/i )  ) {
                     $is_message_moved = 1;
                     last;	# leave attachments check in one message
                  }
               }
            }
            if ( $is_message_moved == 1 ) {
               last;
            }
                        
         } elsif ($rules eq 'attfilename') {
            if ($currmessage eq "") {
               seek($spoolhandle, $attr[$_OFFSET], 0);
               read($spoolhandle, $currmessage, $attr[$_SIZE]);
            }
            if ($header eq "") {
               ($header, $body, $r_attachments)=parse_rfc822block(\$currmessage);
            }

            # check attachments
            foreach my $r_attachment (@{$r_attachments}) {
               if (   ( $include eq 'include' && ${$r_attachment}{filename} =~ /$text/i )
                    ||( $include eq 'exclude' && ${$r_attachment}{filename} !~ /$text/i )  ) {
                  $is_message_moved = 1;
                  last;	# leave attachments check in one message
               }
            }
            if ( $is_message_moved == 1 ) {
               last;
            }
         }
         
      } # end @filterrules
         
      if ($is_message_moved) {
         ## open destination folder, since dest may not exist => open before lock ##
         ($destination =~ /^(.+)$/) && ($destination = $1);  # untaint $destination

         if(! -f "$folderdir/$destination") {
            open (DEST, ">$folderdir/$destination") or
                  neomailerror ("$lang_err{'couldnt_open'} $destination!");
            close (DEST) or neomailerror("$lang_err{'couldnt_close'} $destination!");
            push (@validfolders, $destination);
         }
         open (DEST, ">>$folderdir/$destination") or
               neomailerror ("$lang_err{'couldnt_open'} $destination!");
         unless (filelock("$folderdir/$destination", LOCK_EX|LOCK_NB)) {
            neomailerror("$lang_err{'couldnt_lock'} $destination!");
         }
         update_headerdb("$folderdir/.$destination", "$folderdir/$destination");
             
         filelock("$folderdir/.$destination.$dbm_ext", LOCK_EX);
         dbmopen (%HDB2, "$folderdir/.$destination", 600);

         ## move message ##
         $moved++;

         $messagestart=$attr[$_OFFSET];
         $messagesize=$attr[$_SIZE];

         shiftblock($spoolhandle, $blockstart, $blockend-$blockstart, $writepointer-$blockstart);

         $writepointer=$writepointer+($blockend-$blockstart);
         $blockstart=$blockend=$messagestart+$messagesize;

         if ($currmessage eq "") {
            seek($spoolhandle, $attr[$_OFFSET], 0);
            read($spoolhandle, $currmessage, $attr[$_SIZE]);
         }
            
         $attr[$_OFFSET]=tell(DEST);
         if ($currmessage =~ /^From /) {
            $attr[$_SIZE]=length($currmessage);
            print DEST $currmessage;
         } else {
            $attr[$_SIZE]=length("From ")+length($currmessage);
            print DEST "From ", $currmessage;
         }

         $HDB2{$allmessageids[$i]}=join('@@@', @attr);
         delete $HDB{$allmessageids[$i]};
         if ( $attr[$_STATUS]!~/r/i ) {
            $HDB2{'NEWMESSAGES'}++;
            $HDB{'NEWMESSAGES'}--;
         }
         $HDB2{'ALLMESSAGES'}++;
         $HDB{'ALLMESSAGES'}--;
         
         ## close destination file and dbm ##
         close (DEST);
            
         $HDB2{'METAINFO'}=metainfo("$folderdir/$destination");
         dbmclose(%HDB2);
         filelock("$folderdir/.$destination.$dbm_ext", LOCK_UN);

         filelock("$folderdir/$destination", LOCK_UN);

      } else {
         $messagestart=$attr[$_OFFSET];
         $messagesize=$attr[$_SIZE];
         $blockend=$messagestart+$messagesize;

         my $movement=$writepointer-$blockstart;
         if ($movement<0) {
            $attr[$_OFFSET]+=$movement;
            $HDB{$allmessageids[$i]}=join('@@@', @attr);
         }
      }
   } ## end of allmessages ##

   if ($moved>0) {
      shiftblock($spoolhandle, $blockstart, $blockend-$blockstart, $writepointer-$blockstart);
      seek($spoolhandle, $writepointer+($blockend-$blockstart), 0);
      truncate($spoolhandle, tell($spoolhandle));
   }
   close ($spoolhandle) or neomailerror("$lang_err{'couldnt_close'} $spoolfile!");

   $HDB{'METAINFO'}=metainfo($spoolfile);
   dbmclose(%HDB);
   filelock("$headerdb.$dbm_ext", LOCK_UN);

   filelock($spoolfile, LOCK_UN);
   chown($menu_id,$gid,$spoolfile);
   
   ## update .filter.check ##
   if (-f "$folderdir/.filter.check" ) {
      open (FILTERCHECK, ">$folderdir/.filter.check" ) or
            neomailerror("$lang_err{'couldnt_open'} .filter.check!");
      print FILTERCHECK metainfo($spoolfile);
      truncate(FILTERCHECK, tell(FILTERCHECK));
      close (FILTERCHECK);
   } else {
      open (FILTERCHECK, ">$folderdir/.filter.check" ) or
            neomailerror("$lang_err{'couldnt_open'} .filter.check!");
      print FILTERCHECK metainfo($spoolfile);
      close (FILTERCHECK);
   }
}
################# END FILTERMESSAGE #######################

###################### SETPREFS ###########################
sub setprefs {
	   my $html = '';
	   my $temphtml;

	   open (PREFSTEMPLATE, "$neomaildir/templates/$lang/prefs.template") ||
	      neomailerror("$lang_err{'couldnt_open'} prefs.template");
	   while (<PREFSTEMPLATE>) {
	      $html .= $_;
	   }
	   close (PREFSTEMPLATE);

	   $html = applystyle($html);
	   printheader();
	### Get a list of valid style files
	   opendir (STYLESDIR, $stylesdir) ||
	      neomailerror("$lang_err{'couldnt_open'} $stylesdir directory for reading!");
	   while (defined(my $currentstyle = readdir(STYLESDIR))) {
	      unless ($currentstyle =~ /\./) {
        	 push (@styles, $currentstyle);
	      }
	   }
	   @styles = sort(@styles);
	   closedir(STYLESDIR) || neomailerror("$lang_err{'couldnt_close'} $stylesdir!");

	   $temphtml = '<form method=POST action="'.$cgi_url.'">';
	   $temphtml .= "<input type='hidden' name='step' value='mail'>";
	   $temphtml .= "<input type='hidden' name='action' value='saveprefs'>";
	   $temphtml .= "<input type='hidden' name='firsmessage' value='".$firstmessage."'>";
	   $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
	   $html =~ s/\@\@\@STARTPREFSFORM\@\@\@/$temphtml/;

	   if ($DATA{"realname"}) {
	      $prefs{"realname"} = $DATA{"realname"};
	   }
	   if ($prefs{"realname"}) {
	      $temphtml = " $lang_text{'for'} " . uc($prefs{"realname"});
	   } else {
	      $temphtml = '';
	   }

	   $html =~ s/\@\@\@REALNAME\@\@\@/$temphtml/;

	   $temphtml = "<select name='language'>";
	   foreach (@availablelanguages) {
	      if ($_ eq $prefs{"language"} || $_ eq $defaultlanguage) {
	         $temphtml .= "<option value='".$_."' selected>".$languagenames{$_};
	      } else {
	         $temphtml .= "<option value='".$_."'>".$languagenames{$_};
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@LANGUAGEFIELD\@\@\@/$temphtml/;

	   $temphtml = "<input type='text' name='realname' value='".($prefs{"realname"} || $lang_text{'yourname'})."' size='40'>";
	   $html =~ s/\@\@\@REALNAMEFIELD\@\@\@/$temphtml/;

	   $temphtml = "<input type='text' name='fromname' value='".($prefs{"fromname"} || $UIDNM{$menu_id})."' size='15'>";
	   $html =~ s/\@\@\@USERNAME\@\@\@/$temphtml/;

	   $temphtml = "<select name='domainname'>";
	   foreach (@domainnames) {
	      if ($_ eq $prefs{"domainname"} || $_ eq $domainnames[0]) {
	         $temphtml .= "<option value='".$_."' selected>".$_;
	      } else {
	         $temphtml .= "<option value='".$_."'>".$_;
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@DOMAINFIELD\@\@\@/$temphtml/;

	   $temphtml = "<input type='text' name='replyto' value='".$prefs{"replyto"}."' size='40'>";
	   $html =~ s/\@\@\@REPLYTOFIELD\@\@\@/$temphtml/;

	   $temphtml = "<select name='style'>";
	   foreach (@styles) {
	      if ($_ eq $prefs{"style"} || $_ eq 'Default') {
	         $temphtml .= "<option value='".$_."' selected>".$_;
	      } else {
	         $temphtml .= "<option value='".$_."'>".$_;
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@STYLEMENU\@\@\@/$temphtml/;

	   $temphtml = "<select name='sort'>";
	   foreach ('date','date_rev','sender','sender_rev','size','size_rev','subject','subject_rev','status') {
	      if ($_ eq $prefs{"sort"} || $_ eq 'date') {
	         $temphtml .= "<option value='".$_."' selected>".$lang_sortlabels{$_};
	      } else {
	         $temphtml .= "<option value='".$_."'>".$lang_sortlabels{$_};
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@SORTMENU\@\@\@/$temphtml/;

	   $temphtml = "<select name='numberofmessages'>";
	   foreach ('10','20','30','40','50','100','500','1000') {
	      if ($_ eq $prefs{"numberofmessages"} || $_ eq $numberofheaders) {
	         $temphtml .= "<option value='".$_."' selected>".$_;
	      } else {
	         $temphtml .= "<option value='".$_."'>".$_;
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@NUMBEROFMESSAGES\@\@\@/$temphtml/;

	   my %headerlabels = ('simple'=>$lang_text{'simplehead'},
        	               'all'=>$lang_text{'allhead'} );
	   $temphtml = "<select name='headers'>";
	   foreach ('simple','all') {
	      if ($_ eq $prefs{"headers"} || $_ eq 'simple') {
	         $temphtml .= "<option value='".$_."' selected>".$headerlabels{$_};
	      } else {
	         $temphtml .= "<option value='".$_."'>".$headerlabels{$_};
	      }
	   }
	   $temphtml .= "</select>";
	   $html =~ s/\@\@\@HEADERSMENU\@\@\@/$temphtml/;

	   unless (defined($prefs{"signature"})) {
	      $prefs{"signature"} = $defaultsignature;
	   }
	   $temphtml = "<textarea name='signature' rows='5' cols='72'>".$prefs{"signature"}."</textarea>";
	   $html =~ s/\@\@\@SIGAREA\@\@\@/$temphtml/;

	   $temphtml = "<input type='submit' name='".$lang_text{'save'}."' value='".$lang_text{'save'}."'></form>";

	   unless ( $firsttimeuser eq 'yes' ) {
	      $temphtml .= '<form method=POST action="'.$cgi_url.'">';
	      $temphtml .= "<input type='hidden' name='step' value='mail'>";
	      $temphtml .= "<input type='hidden' name='action' value='displayheaders'>";
   	      $temphtml .= "<input type='hidden' name='firsmessage' value='".$firstmessage."'>";
	      $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
	      $temphtml .= "<input type='hidden' name='folder' value='".$folder."'></td><td>";
	      $temphtml .= "<input type='submit' name='".$lang_text{'cancel'}."' value='".$lang_text{'cancel'}."'></form>";
	}
	$html =~ s/\@\@\@BUTTONS\@\@\@/$temphtml/;
	print $html;
	printfooter();
     }
#################### END SETPREFS ##########################

#################### EDITFOLDERS ###########################
sub editfolders {
   my @folders;
   opendir (FOLDERDIR, "$folderdir") or
      neomailerror("$lang_err{'couldnt_open'} $folderdir!");
   while (defined(my $filename = readdir(FOLDERDIR))) {
### skip files started with ., which are neomail internal files (dbm, search caches)
      if ( $filename=~/^\./ ) {	
         next;
      }

      if ($homedirfolders eq 'yes') {
         unless ( ($filename eq 'saved_messages') ||
                  ($filename eq 'sent_mail') ||
                  ($filename eq 'neomail_trash') ||
                  ($filename eq '.') ||
                  ($filename eq '..')
                ) {
            push (@folders, $filename);
         }
      } else {
         if ($filename =~ /^(.+)\.folder$/) {
            push (@folders, $1);
         }
      }
   }
   closedir (FOLDERDIR) or
      neomailerror("$lang_err{'couldnt_close'} $folderdir!");

   my $html = '';
   my $temphtml;

   open (EDITFOLDERSTEMPLATE, "$neomaildir/templates/$lang/editfolders.template") or
      neomailerror("$lang_err{'couldnt_open'} editfolders.template!");
   while (<EDITFOLDERSTEMPLATE>) {
      $html .= $_;
   }
   close (EDITFOLDERSTEMPLATE);

   $html = applystyle($html);

   printheader();

   $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$escapedfolder\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a>";

   $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

   $temphtml = '<form method=POST action="'.$cgi_url.'">';
   $temphtml .= "<input type='hidden' name='step' value='mail'>";
   $temphtml .= "<input type='hidden' name='action' value='addfolder'>";
   $temphtml .= "<input type='hidden' name='firsmessage' value='".$firstmessage."'>";
   $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
   $temphtml .= "<input type='hidden' name='folder' value='".$folder."'></td><td>";
   $html =~ s/\@\@\@STARTFOLDERFORM\@\@\@/$temphtml/;

   $temphtml = "<input type='text' name='foldername' size=16 maxlength=16>";
   $html =~ s/\@\@\@FOLDERNAMEFIELD\@\@\@/$temphtml/;

   $temphtml = "<input type='submit' name='".$lang_text{'add'}."' value='".$lang_text{'add'}."'>";
   $html =~ s/\@\@\@ADDBUTTON\@\@\@/$temphtml/;

   $temphtml = "</form>";
   $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/;

   $temphtml = '';
   my $bgcolor = $style{"tablerow_dark"};
   my $currfolder;
   foreach $currfolder (sort (@folders)) {
      $temphtml .= "<tr><td align=\"center\" bgcolor=$bgcolor>$currfolder</td>";
      $temphtml .= "<form method=POST action=\"".$cgi_url."\" onsubmit=\"return confirm('".$lang_text{'folderconf'}."')\">";
      $temphtml .= "<input type='hidden' name='step' value='mail'>";
      $temphtml .= "<input type='hidden' name='action' value='deletefolder'>";
      $temphtml .= "<input type='hidden' name='firsmessage' value='".$firstmessage."'>";
      $temphtml .= "<input type='hidden' name='sort' value='".$sort."'>";
      $temphtml .= "<input type='hidden' name='folder' value='".$folder."'>";
      $temphtml .= "<input type='hidden' name='foldername' value='".$currfolder."'>";
      $temphtml .= "<td bgcolor=$bgcolor align=\"center\">";
      $temphtml .= "<input type=submit name='".$lang_text{'delete'}."' value='".$lang_text{'delete'}."'>";
      $temphtml .= '</td></tr></form>';
      if ($bgcolor eq $style{"tablerow_dark"}) {
         $bgcolor = $style{"tablerow_light"};
      } else {
         $bgcolor = $style{"tablerow_dark"};
      }
   }

   $html =~ s/\@\@\@FOLDERS\@\@\@/$temphtml/;
   print $html;

   printfooter();
}
################### END EDITFOLDERS ########################

################### ADDFOLDER ##############################
sub addfolder {
   my $foldertoadd = $DATA{'foldername'} || '';
   $foldertoadd =~ s/[\.\.|\/|\\|\`|;|<|>]//g;
   unless ($homedirfolders eq 'yes') {
      $foldertoadd = uc($foldertoadd);
   }
   if (length($foldertoadd) > 16) {
      neomailerror("$lang_err{'foldername_long'}");
   }
   ($foldertoadd =~ /^(.+)$/) && ($foldertoadd = $1);
   unless ($homedirfolders eq 'yes') {
      $foldertoadd .= '.folder';
   }
   if ( -f "$folderdir/$foldertoadd" ) {
      neomailerror ("$lang_err{'folder_with_name'} $foldertoadd $lang_err{'already_exists'}");
   }
#   if ( ($homedirfolders eq 'yes') && ($> == 0) ) {
#      $) = $gid;
#      $> = $menu_id;
#      umask(0077); # make sure only owner can read/write
#   }
   open (FOLDERTOADD, ">$folderdir/$foldertoadd") or
      neomailerror("$lang_err{'cant_create_folder'}");
   close (FOLDERTOADD) or neomailerror("$lang_err{'couldnt_close'} $foldertoadd!");
   print "Location: $scripturl&action=editfolders&sort=$sort&folder=$escapedfolder&firstmessage=$firstmessage\n\n";
}
################### END ADDFOLDER ##########################

################### DELETEFOLDER ##############################
sub deletefolder {
   my $foldertodel = $DATA{'foldername'} || '';
   $foldertodel =~ s/\.\.+//g;
   $foldertodel =~ s/[\/|\\|\`|;|<|>]//g;
   ($foldertodel =~ /^(.+)$/) && ($foldertodel = $1);
   unless ($homedirfolders eq 'yes') {
      $foldertodel .= '.folder';
   }
   if ( -f "$folderdir/$foldertodel" ) {
      unlink ("$folderdir/$foldertodel",
              "$folderdir/.$foldertodel.db",
	      "$folderdir/.$foldertodel.dir",
              "$folderdir/.$foldertodel.pag",
              "$folderdir/.$foldertodel.cache");              
   }
   print "Location: $scripturl&action=editfolders&sort&folder=$escapedfolder&firstmessage=$firstmessage\n\n";
}
################### END DELETEFOLDER ##########################

##################### IMPORTABOOK ############################
sub importabook {
   my ($name, $email);
   my %addresses;
   my $abookcontents = $DATA{"abook"} || '';
   my $abooktowrite='';
   my $mua = $DATA{"mua"} || '';
   if ($abookcontents) {
      unless ( -f "$folderdir/.address.book" ) {
         open (ABOOK, ">>$folderdir/.address.book"); # Create if nonexistent
         close(ABOOK);
      }
      open (ABOOK,"+<$folderdir/.address.book") or
         neomailerror("$lang_err{'couldnt_open'} .address.book!");
      unless (flock(ABOOK, LOCK_EX|LOCK_NB)) {
         neomailerror("$lang_err{'couldnt_lock'} .address.book!");
      }
      while (<ABOOK>) {
         ($name, $email) = split(/:/, $_);
         chomp($email);
         $addresses{"$name"} = $email;
      }
      my @fields;
      my $quotecount;
      my $tempstr;
      my @processed;
      my $flag = 1;

      foreach (split(/\r*\n/, $abookcontents)) {
         if ( ($mua eq 'outlookexp5') && ( $flag==1) ) { 
	    $flag = 0;
	    next; 
	 }
         next if ( $_ !~ (/\@/) );
         $quotecount = 0;
         @fields = split(/,/);
         @processed = ();
         $tempstr = '';
         foreach my $str (@fields) {
            if ( ($str =~ /"/) && ($quotecount == 1) ) {
               $tempstr .= ',' . $str;
               $tempstr =~ s/"//g;
               push (@processed, $tempstr);
               $tempstr = '';
               $quotecount = 0;
            } elsif ($str =~ /"/) {
               $tempstr .= $str;
               $quotecount = 1;
            } elsif ($quotecount == 1) {
               $tempstr .= ',' . $str;
            } else {
               push (@processed, $str);
            }
         }
         if ( ($mua eq 'outlookexp5') && ($processed[0]) && ($processed[1]) ) {
            $processed[0] =~ s/^\s*//;
            $processed[0] =~ s/\s*$//;
            $processed[0] =~ s/://;
            $processed[0] =~ s/</&lt;/g;
            $processed[0] =~ s/>/&gt;/g;
            $processed[1] =~ s/</&lt;/g;
            $processed[1] =~ s/>/&gt;/g;
            $addresses{"$processed[0]"} = $processed[1];
         } elsif ( ($mua eq 'nsmail') && ($processed[0]) && ($processed[6]) ) {
            $processed[0] =~ s/^\s*//;
            $processed[0] =~ s/\s*$//;
            $processed[0] =~ s/://;
            $processed[0] =~ s/</&lt;/g;
            $processed[0] =~ s/>/&gt;/g;
            $processed[6] =~ s/</&lt;/g;
            $processed[6] =~ s/>/&gt;/g;
            $addresses{"$processed[0]"} = $processed[6];
         }
      }

      seek (ABOOK, 0, 0) or
         neomailerror("$lang_err{'couldnt_seek'} .address.book!");

      while ( ($name, $email) = each %addresses ) {
         $abooktowrite .= "$name:$email\n";
      }

      if (length($abooktowrite) > ($maxabooksize * 1024)) {
         neomailerror("$lang_err{'abook_toobig'}
                       <a href=\"$scripturl&amp;action=importabook&amp;sort=$sort&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;message_id=$escapedmessageid\">$lang_err{'back'}</a>
                       $lang_err{'tryagain'}");
      }
      print ABOOK $abooktowrite;
      truncate(ABOOK, tell(ABOOK));
      close (ABOOK) or neomailerror("$lang_err{'couldnt_close'} .address.book!");

      print "Location: $scripturl&action=editaddresses&sort&folder=$escapedfolder&firstmessage=$firstmessage&message_id=$escapedmessageid\n\n";

   } else {

      my $abooksize = ( -s "$folderdir/.address.book" ) || 0;
      my $freespace = int($maxabooksize - ($abooksize/1024) + .5);

      my $html = '';
      my $temphtml;

      open (IMPORTTEMPLATE, "$neomaildir/templates/$lang/importabook.template") or
         neomailerror("$lang_err{'couldnt_open'} importabook.template");
      while (<IMPORTTEMPLATE>) {
         $html .= $_;
      }
      close (IMPORTTEMPLATE);

      $html = applystyle($html);

      printheader();

      $html =~ s/\@\@\@FREESPACE\@\@\@/$freespace/g;

      $temphtml = '<FORM METHOD="POST" ENCTYPE="multipart/form-data">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="importabook">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
      $html =~ s/\@\@\@STARTIMPORTFORM\@\@\@/$temphtml/;

      $temphtml = "<input type=radio name='mua' value='outlookexp5' checked>".$lang_mualabels{'outlookexp5'};
      $temphtml .= "<input type=radio name='mua' value='nsmail'>".$lang_mualabels{'nsmail'};
      $html =~ s/\@\@\@MUARADIOGROUP\@\@\@/$temphtml/;

      $temphtml = '<INPUT TYPE="file" NAME="abook" size=30>';
      $html =~ s/\@\@\@IMPORTFILEFIELD\@\@\@/$temphtml/;

      $temphtml = "<input type=submit name='".$lang_text{'import'}."' value='".$lang_text{'import'}."'>";
      $html =~ s/\@\@\@IMPORTBUTTON\@\@\@/$temphtml/;

      $temphtml = "</form>";
      $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/g;

      $temphtml = '<FORM METHOD="POST" action="'.$cgi_url.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="editaddresses">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
      $html =~ s/\@\@\@STARTCANCELFORM\@\@\@/$temphtml/;


      $temphtml = "<input type=submit name='".$lang_text{'cancel'}."' value='".$lang_text{'cancel'}."'>";
      $html =~ s/\@\@\@CANCELBUTTON\@\@\@/$temphtml/;

      print $html;

      printfooter();
   }
}
#################### END IMPORTABOOK #########################

#################### EDITADDRESSES ###########################
sub editaddresses {
   my %addresses;
   my ($name, $email);

   my $html = '';
   my $temphtml;

   open (EDITABOOKTEMPLATE, "$neomaildir/templates/$lang/editaddresses.template") or
      neomailerror("$lang_err{'couldnt_open'} editaddresses.template");
   while (<EDITABOOKTEMPLATE>) {
      $html .= $_;
   }
   close (EDITABOOKTEMPLATE);

   $html = applystyle($html);

   if ( -f "$folderdir/.address.book" ) {
      open (ABOOK,"$folderdir/.address.book") or
         neomailerror("$lang_err{'couldnt_open'} .address.book!");
      while (<ABOOK>) {
         ($name, $email) = split(/:/, $_);
         chomp($email);
         $addresses{"$name"} = $email;
      }
      close (ABOOK) or neomailerror("$lang_err{'couldnt_close'} .address.book!");
   }
   my $abooksize = ( -s "$folderdir/.address.book" ) || 0;
   my $freespace = int($maxabooksize - ($abooksize/1024) + .5);

   printheader();

   $html =~ s/\@\@\@FREESPACE\@\@\@/$freespace/g;

   if ( $DATA{"message_id"} ) {
      $temphtml = "<a href=\"$scripturl&amp;action=readmessage&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a> &nbsp; &nbsp; ";
   } else {
      $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a> &nbsp; &nbsp; ";
   }
   $temphtml .= "<a href=\"$scripturl&amp;action=importabook&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/import.gif\" border=\"0\" ALT=\"$lang_text{'importadd'}\"></a>";

   $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

   $temphtml = '<FORM METHOD="POST" action="'.$cgi_url.'" name="newaddress">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="addaddress">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
   $html =~ s/\@\@\@STARTADDRESSFORM\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="text" NAME="realname" size=25>';
   $html =~ s/\@\@\@REALNAMEFIELD\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="text" NAME="email" size=35>';
   $html =~ s/\@\@\@EMAILFIELD\@\@\@/$temphtml/;

   $temphtml = "<input type=submit name='".$lang_text{'addmod'}."' value='".$lang_text{'addmod'}."'>";
   $html =~ s/\@\@\@ADDBUTTON\@\@\@/$temphtml/;

   $temphtml = "</form>";
   $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/;

   $temphtml = '';
   my $bgcolor = $style{"tablerow_dark"};
   foreach my $key (sort { uc($a) cmp uc($b) } (keys %addresses)) {
      $temphtml .= "<tr><td bgcolor=$bgcolor width=\"200\">
                    <a href=\"Javascript:Update('$key','$addresses{$key}')\">
                    $key</a></td><td bgcolor=$bgcolor width=\"200\">
                    <a href=\"$scripturl&amp;action=composemessage&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;folder=$escapedfolder&amp;composetype=sendto&amp;to=$addresses{$key}\">$addresses{$key}</a></td>";

      $temphtml .= '<FORM METHOD="POST" action="'.$cgi_url.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="deleteaddress">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="realname" VALUE="'.$key.'">';
      $temphtml .= "<td bgcolor=$bgcolor align=\"center\" width=\"100\">";
      $temphtml .= "<input type=submit name='".$lang_text{'delete'}."' value='".$lang_text{'delete'}."'>";
      $temphtml .= '</td></tr></form>';
      if ($bgcolor eq $style{"tablerow_dark"}) {
         $bgcolor = $style{"tablerow_light"};
      } else {
         $bgcolor = $style{"tablerow_dark"};
      }
   }

   $html =~ s/\@\@\@ADDRESSES\@\@\@/$temphtml/;
   print $html;

   printfooter();
}
################### END EDITADDRESSES ########################

################### MODADDRESS ##############################
sub modaddress {
   my $mode = shift;
   my ($realname, $address);
   $realname = $DATA{"realname"} || '';
   $address = $DATA{"email"} || '';
   $realname =~ s/://;
   $realname =~ s/^\s*//; # strip beginning and trailing spaces from hash key
   $realname =~ s/\s*$//;
   $address =~ s/[#&=\?]//g;

   if (($realname && $address) || (($mode eq 'delete') && $realname) ) {
      my %addresses;
      my ($name,$email);
      if ( -f "$folderdir/.address.book" ) {
         my $abooksize = ( -s "$folderdir/.address.book" );
         if ( (($abooksize + length($realname) + length($address) + 2) >= ($maxabooksize * 1024) ) && ($mode ne "delete") ) {
            neomailerror("$lang_err{'abook_toobig'} <a href=\"$scripturl&action=editaddresses&amp;sort=$sort&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;message_id=$escapedmessageid\">$lang_err{'back'}</a>
                          $lang_err{'tryagain'}");
         }
         open (ABOOK,"+<$folderdir/.address.book") or
            neomailerror("$lang_err{'couldnt_open'} .address.book!");
         unless (flock(ABOOK, LOCK_EX|LOCK_NB)) {
            neomailerror("$lang_err{'couldnt_lock'} .address.book!");
         }
         while (<ABOOK>) {
            ($name, $email) = split(/:/, $_);
            chomp($email);
            $addresses{"$name"} = $email;
         }
         if ($mode eq 'delete') {
            delete $addresses{"$realname"};
         } else {
            $addresses{"$realname"} = $address;
         }
         seek (ABOOK, 0, 0) or
            neomailerror("$lang_err{'couldnt_seek'} .address.book!");
         while ( ($name, $email) = each %addresses ) {
            print ABOOK "$name:$email\n";
         }
         truncate(ABOOK, tell(ABOOK));
         close (ABOOK) or neomailerror("$lang_err{'couldnt_close'} .address.book!");
      } else {
         open (ABOOK, ">$folderdir/.address.book" ) or
            neomailerror("$lang_err{'couldnt_open'} .address.book!");
         print ABOOK "$realname:$address\n";
         close (ABOOK) or neomailerror("$lang_err{'couldnt_close'} .address.book!");
      }
   }

   print "Location: $scripturl&action=editaddresses&sort=$sort&folder=$escapedfolder&firstmessage=$firstmessage&amp;message_id=$escapedmessageid\n\n";
}
################## END MODADDRESS ###########################

#################### EDITPOP3 ###########################
sub editpop3 {
   my %account;
   my ($name, $pass, $host, $del);

   my $html = '';
   my $temphtml;

   open (EDITPOP3BOOKTEMPLATE, "$neomaildir/templates/$lang/editpop3.template") or
      neomailerror("$lang_err{'couldnt_open'} editpop3.template");
   while (<EDITPOP3BOOKTEMPLATE>) {
      $html .= $_;
   }
   close (EDITPOP3BOOKTEMPLATE);

   $html = applystyle($html);

   if ( -f "$folderdir/.pop3.book" ) {
      open (POP3BOOK,"$folderdir/.pop3.book") or
         neomailerror("$lang_err{'couldnt_open'} .pop3.book!");
      while (<POP3BOOK>) {
      	 chomp($_);
         ($host, $name, $pass, $del) = split(/:/, $_);
         $account{"$host:$name"} = "$host:$name:$pass:$del";
      }
      close (POP3BOOK) or neomailerror("$lang_err{'couldnt_close'} .pop3.book!");
   }
   my $abooksize = ( -s "$folderdir/.pop3.book" ) || 0;
   my $freespace = int($maxabooksize - ($abooksize/1024) + .5);

   printheader();

   $html =~ s/\@\@\@FREESPACE\@\@\@/$freespace/g;

   if ( $DATA{"message_id"} ) {
      $temphtml = "<a href=\"$scripturl&amp;action=readmessage&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a> &nbsp; &nbsp; ";
   } else {
      $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a> &nbsp; &nbsp; ";
   }
   $temphtml .= "<a href=\"$scripturl&amp;action=retrpop3s&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/pop3.gif\" border=\"0\" ALT=\"$lang_text{'retr_pop3s'}\"></a>";

   $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

   $temphtml = '<FORM METHOD="POST" action="'.$cgi_url.'" name="newpop3">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="addpop3">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
   $html =~ s/\@\@\@STARTADDRESSFORM\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="text" NAME="host" size=32>';
   $html =~ s/\@\@\@HOSTFIELD\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="text" NAME="name" size=16>';
   $html =~ s/\@\@\@REALNAMEFIELD\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="password" NAME="pass" size=12>';
   $html =~ s/\@\@\@PASSFIELD\@\@\@/$temphtml/;

   $temphtml = '<INPUT TYPE="checkbox" NAME="del" value="1">';
   $html =~ s/\@\@\@DELFIELD\@\@\@/$temphtml/;

   $temphtml = "<input type=submit name='".$lang_text{'addmod'}."' value='".$lang_text{'addmod'}."'>";
   $html =~ s/\@\@\@ADDBUTTON\@\@\@/$temphtml/;

   $temphtml = "</form>";
   $html =~ s/\@\@\@ENDFORM\@\@\@/$temphtml/;

   $temphtml = '';
   my $bgcolor = $style{"tablerow_dark"};
   #foreach my $key (sort { uc($a) cmp uc($b) } (keys %account)) 
   foreach (values %account) {
      ($host, $name, $pass, $del) = split(/:/, $_);
      #<td bgcolor=$bgcolor width=\"200\"><a href=\"$scripturl&action=retrpop3&name=$name&host=$host&del=0&mbox=$name\@$host&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;folder=$escapedfolder&amp;\">$host</a></td>
      $temphtml .= "<tr>
                    <td bgcolor=$bgcolor width=\"200\"><a href=\"$scripturl&action=retrpop3&name=$name&host=$host&amp;firstmessage=$firstmessage&amp;sort=$sort&amp;folder=$escapedfolder&amp;\">$host</a></td>
      		    <td align=\"center\" bgcolor=$bgcolor width=\"200\"><a href=\"Javascript:Update('$name','$pass','$host','$del')\">$name</a></td>
                    <td align=\"center\" bgcolor=$bgcolor width=\"200\">\*\*\*\*\*\*</td>
                    <td align=\"center\" bgcolor=$bgcolor width=\"200\">";
      if ( $del == 1) {
      	 $temphtml .= $lang_text{'delete'};
      }
      else {
      	 $temphtml .= $lang_text{'reserve'};
      }
      $temphtml .= "</td>";

      $temphtml .= '<FORM METHOD="POST" action="'.$cgi_url.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="deletepop3">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="name" VALUE="'.$name.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="host" VALUE="'.$host.'">';
      $temphtml .= "<td bgcolor=$bgcolor align=\"center\" width=\"100\">";
      $temphtml .= "<input type=submit name='".$lang_text{'delete'}."' value='".$lang_text{'delete'}."'>";
      $temphtml .= '</td></tr>';
      $temphtml .= "</form>";
      if ($bgcolor eq $style{"tablerow_dark"}) {
         $bgcolor = $style{"tablerow_light"};
      } else {
         $bgcolor = $style{"tablerow_dark"};
      }
   }

   $html =~ s/\@\@\@ADDRESSES\@\@\@/$temphtml/;
   print $html;

   printfooter();
}
################### END EDITPOP3 ########################

################### MODPOP3 ##############################
sub modpop3 {
   my $mode = shift;
   my ($host, $name, $pass, $del, $lastid);
   $host = $DATA{"host"} || '';
   $name = $DATA{"name"} || '';
   $pass = $DATA{"pass"} || '';
   $del = $DATA{"del"} || 0;
   $lastid = "none";
   
   # strip beginning and trailing spaces from hash key
   $host =~ s/://;
   $host =~ s/^\s*//; 
   $host =~ s/\s*$//;
   $host =~ s/[#&=\?]//g;
   
   $name =~ s/://;
   $name =~ s/^\s*//; 
   $name =~ s/\s*$//;
   $name =~ s/[#&=\?]//g;
   
   $pass =~ s/://;
   $pass =~ s/^\s*//; 
   $pass =~ s/\s*$//;

   if (($host && $name && $pass) || (($mode eq 'delete') && $host && $name) ) {
      my %account;
      
      if ( -f "$folderdir/.pop3.book" ) {
         my $pop3booksize = ( -s "$folderdir/.pop3.book" );
         if ( (($pop3booksize + length($host) + length($name)+ length($pass)+ 3) >= ($maxabooksize * 1024) ) && ($mode ne "delete") ) {
            neomailerror("$lang_err{'abook_toobig'} <a href=\"$scripturl&amp;action=editpop3&amp;sort=$sort&amp;folder=$escapedfolder&amp;firstmessage=$firstmessage&amp;message_id=$escapedmessageid\">$lang_err{'back'}</a>
                          $lang_err{'tryagain'}");
         }
         open (POP3BOOK,"+<$folderdir/.pop3.book") or
            neomailerror("$lang_err{'couldnt_open'} .pop3.book!");
         unless (flock(POP3BOOK, LOCK_EX|LOCK_NB)) {
            neomailerror("$lang_err{'couldnt_lock'} .pop3.book!");
         }
         while (<POP3BOOK>) {
         	my ($ehost,$ename,$epass,$edel,$elastid);
         	chomp($_);
            ($ehost, $ename, $epass, $edel, $elastid) = split(/:/, $_);
            $account{"$ehost:$ename"}="$ehost:$ename:$epass:$edel:$elastid";
         }
         if ($mode eq 'delete') {
            delete $account{"$host:$name"};
         } else {
            $account{"$host:$name"}="$host:$name:$pass:$del:$lastid";
         }
         seek (POP3BOOK, 0, 0) or
            neomailerror("$lang_err{'couldnt_seek'} .pop3.book!");
         	
         foreach (values %account) {
            print POP3BOOK "$_\n";
         }
         truncate(POP3BOOK, tell(POP3BOOK));
         close (POP3BOOK) or neomailerror("$lang_err{'couldnt_close'} .pop3.book!");
      } else {
         open (POP3BOOK, ">$folderdir/.pop3.book" ) or
            neomailerror("$lang_err{'couldnt_open'} .pop3.book!");
         print POP3BOOK "$host:$name:$pass:$del:$lastid\n";
         close (POP3BOOK) or neomailerror("$lang_err{'couldnt_close'} .pop3.book!");
      }
   }

   print "Location: $scripturl&action=editpop3&sort=$sort&folder=$escapedfolder&firstmessage=$firstmessage&amp;message_id=$escapedmessageid\n\n";
}
################## END MODPOP3 ###########################

#################### EDITFILTER ###########################
sub editfilter {
   #### variables ####
   my $html = '';
   my $temphtml;
   my @validfolders;
   my @filterrules;

   #### UI: header ####
   printheader();

   #### UI: between header/footer ####
   open (EDITFILTERTEMPLATE, "$neomaildir/templates/$lang/editfilter.template") or
       neomailerror("$lang_err{'couldnt_open'} editfilter.template");
   while (<EDITFILTERTEMPLATE>) {
       $html .= $_;
   }
   close (EDITFILTERTEMPLATE);

   ## user-prefer-UI-style ##
   $html = applystyle($html);
   
   ## replace @@@FREESPACE@@@ ##
   my $abooksize = ( -s "$folderdir/.filter.book" ) || 0;
   my $freespace = int($maxabooksize - ($abooksize/1024) + .5);
   $html =~ s/\@\@\@FREESPACE\@\@\@/$freespace/g;
   
   ## replace @@@MENUBARLINKS@@@ ##
   if ( $DATA{"message_id"} ) {
      $temphtml = "<a href=\"$scripturl&amp;action=readmessage&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder&amp;message_id=$escapedmessageid\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a>";
   } else {
      $temphtml = "<a href=\"$scripturl&amp;action=displayheaders&amp;sort=$sort&amp;firstmessage=$firstmessage&amp;folder=$folder\"><IMG SRC=\"$image_url/backtofolder.gif\" border=\"0\" ALT=\"$lang_text{'backto'} $printfolder\"></a>";
   }
   $html =~ s/\@\@\@MENUBARLINKS\@\@\@/$temphtml/g;

   ## replace @@@STARTFILTERFORM@@@ ##
   $temphtml = '<FORM METHOD="POST" action="'.$cgi_url.'" name="newfilter">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="addfilter">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
   $html =~ s/\@\@\@STARTFILTERFORM\@\@\@/$temphtml/;

   ## replace @@@ENDFILTERFORM@@@ ##
   $temphtml = "</form>";
   $html =~ s/\@\@\@ENDFILTERFORM\@\@\@/$temphtml/;

   ## replace @@@PRIORITYMENU@@@ ##
   $temphtml = "<select name='priority'>";
   foreach ('01','02','03','04','05','06','07','08','09','10','11','12','13','14','15','16','17','18','19','20') {
      if ($_ eq '10') {
         $temphtml .= "<option value='".$_."' selected>".$_;
      } else {
         $temphtml .= "<option value='".$_."'>".$_;
      }
   }
   $temphtml .= "</select>";
   $html =~ s/\@\@\@PRIORITYMENU\@\@\@/$temphtml/;

   ## replace @@@RULEMENU@@@ ##
   %labels = ('from'=>$lang_text{'from'},
                        'to'=>$lang_text{'to'},
                        'subject'=>$lang_text{'subject'},
                        'header'=>$lang_text{'header'},
                        'body'=>$lang_text{'body'},
                        'attfilename'=>$lang_text{'attfilename'});
   $temphtml = "<select name='rules'>";
   foreach ('from', 'to', 'subject', 'header', 'body' ,'attfilename') {
         $temphtml .= "<option value='".$_."'>".$labels{$_};
   }
   $temphtml .= "</select>";
   $html =~ s/\@\@\@RULEMENU\@\@\@/$temphtml/;

   ## replace @@@INCLUDEMENU@@@ ##
   my %labels = ('include'=>$lang_text{'include'},
                        'exclude'=>$lang_text{'exclude'});
   $temphtml = "<select name='include'>";
   foreach ('include', 'exclude') {
         $temphtml .= "<option value='".$_."'>".$labels{$_};
   }
   $temphtml .= "</select>";
   $html =~ s/\@\@\@INCLUDEMENU\@\@\@/$temphtml/;

   ## replace @@@TEXTFIELD@@@ ##
   $temphtml = '<INPUT TYPE="text" NAME="text" size=26>';
   $html =~ s/\@\@\@TEXTFIELD\@\@\@/$temphtml/;

   ## replace @@@FOLDERMENU@@@ ##
   @validfolders = @{&getfolders()};
   foreach (@validfolders) {
      if ( defined($lang_folders{$_}) ) {
          $labels{$_} = $lang_folders{$_};
      } else { 
         $labels{$_} = $_; 
      }
   }
   $temphtml = "<select name='destination'>";
   foreach (@validfolders) {
         $temphtml .= "<option value='".$_."'>".$labels{$_};
   }
   $temphtml .= "</select>";
   $html =~ s/\@\@\@FOLDERMENU\@\@\@/$temphtml/;
   
   ## replace @@@ENABLECHECKBOX@@@ ##
   $temphtml = '<INPUT TYPE="checkbox" NAME="enable" value="1" checked>';
   $html =~ s/\@\@\@ENABLECHECKBOX\@\@\@/$temphtml/;
   
   ## replace @@@ADDBUTTON@@@ ##
   $temphtml = '<INPUT TYPE="submit" NAME="'.$lang_text{'addmod'}.'" value="'.$lang_text{'addmod'}.'">';
   $html =~ s/\@\@\@ADDBUTTON\@\@\@/$temphtml/;

   ## replace @@@FILTERRULES@@@ ##
   if ( -f "$folderdir/.filter.book" ) {
      open (FILTER,"$folderdir/.filter.book") or
         neomailerror("$lang_err{'couldnt_open'} .filter.book!");
      while (<FILTER>) {
          chomp($_);
          push (@filterrules, $_);
      }
      close (FILTER) or neomailerror("$lang_err{'couldnt_close'} .filter.book!");
   }
   $temphtml = '';
   my $bgcolor = $style{"tablerow_dark"};
   foreach (@filterrules) {
      my ($priority, $rules, $include, $text, $destination, $enable) = split(/\@\@\@/, $_);
            
      $temphtml .= "<tr>
               <td bgcolor=$bgcolor align=center>$priority</td>
               <td bgcolor=$bgcolor align=center>$lang_text{$rules}</td>
               <td bgcolor=$bgcolor align=center>$lang_text{$include}</td>
               <td bgcolor=$bgcolor align=center><a href=\"Javascript:Update('$priority','$rules','$include','$text','$destination','$enable')\">$text</a></td>";
      if (defined($lang_folders{$destination})) {
         $temphtml .= "<td bgcolor=$bgcolor align=center>$lang_folders{$destination}</td>";
      } else {
         $temphtml .= "<td bgcolor=$bgcolor align=center>$destination</td>";
      }
      if ($enable == 1) {
         $temphtml .= "<td bgcolor=$bgcolor align=center>$lang_text{'enable'}</td>";
      } else {
         $temphtml .= "<td bgcolor=$bgcolor align=center>$lang_text{'disable'}</td>";
      }
      $temphtml .= '<FORM METHOD="POST" action="'.$cgi_url.'">';
      $temphtml .= "<td bgcolor=$bgcolor align=center>";
      $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="deletefilter">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="message_id" VALUE="'.$message_id.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="rules" VALUE="'.$rules.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="include" VALUE="'.$include.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="text" VALUE="'.$text.'">';
      $temphtml .= '<INPUT TYPE="hidden" NAME="destination" VALUE="'.$destination.'">';
      $temphtml .= '<INPUT TYPE="submit" NAME="'.$lang_text{'delete'}.'" value="'.$lang_text{'delete'}.'">';
      $temphtml .= "</td></tr></form>";
      if ($bgcolor eq $style{"tablerow_dark"}) {
         $bgcolor = $style{"tablerow_light"};
      } else {
         $bgcolor = $style{"tablerow_dark"};
      }
   }
   $html =~ s/\@\@\@FILTERRULES\@\@\@/$temphtml/;

   print $html;
   
   printfooter();
}
################### END EDITFILTER ########################

################### MODFILTER ##############################
sub modfilter {
   ## get parameters ##
   my $mode = shift;
   my ($priority, $rules, $include, $text, $destination, $enable);
   $priority = $DATA{"priority"} || '';
   $rules = $DATA{"rules"} || '';
   $include = $DATA{"include"} || '';
   $text = $DATA{"text"} || '';
   $text =~ s/^[\s\@]+//;
   $text =~ s/[\@\s]+$//;
   $text =~ s/\@\@\@//g;
   $destination = $DATA{"destination"} || '';
   $enable = $DATA{"enable"} || 0;
   
   ## add mode -> can't have null $rules, null $text, null $destination ##
   ## delete mode -> can't have null $filter ##
   if(($rules && $include && $text && $destination && $priority) || (($mode eq 'delete') && ($rules && $include && $text && $destination))) {
      my %filterrules;
      if ( -f "$folderdir/.filter.book" ) {
         open (FILTER,"+<$folderdir/.filter.book") or
               neomailerror("$lang_err{'couldnt_open'} .filter.book!");
         unless (flock(FILTER, LOCK_EX|LOCK_NB)) {
            neomailerror("$lang_err{'couldnt_lock'} .filter.book!");
         }
         while(<FILTER>) {
            my ($epriority,$erules,$einclude,$etext,$edestination,$eenable);
            chomp($_);
            ($epriority,$erules,$einclude,$etext,$edestination,$eenable) = split(/\@\@\@/, $_);
            $filterrules{"$erules\@\@\@$einclude\@\@\@$etext\@\@\@$edestination"}="$epriority\@\@\@$erules\@\@\@$einclude\@\@\@$etext\@\@\@$edestination\@\@\@$eenable";
         }
         if ($mode eq 'delete') {
            delete $filterrules{"$rules\@\@\@$include\@\@\@$text\@\@\@$destination"};
         } else {
            $filterrules{"$rules\@\@\@$include\@\@\@$text\@\@\@$destination"}="$priority\@\@\@$rules\@\@\@$include\@\@\@$text\@\@\@$destination\@\@\@$enable";
         }
         seek (FILTER, 0, 0) or
            neomailerror("$lang_err{'couldnt_seek'} .filter.book!");
   
         foreach (sort values %filterrules) {
            print FILTER "$_\n";
         }
         truncate(FILTER, tell(FILTER));
         close (FILTER) or neomailerror("$lang_err{'couldnt_close'} .filter.book!");         
      } else {
         open (FILTER, ">$folderdir/.filter.book" ) or
                  neomailerror("$lang_err{'couldnt_open'} .filter.book!");
         print FILTER "$priority\@\@\@$rules\@\@\@$include\@\@\@$text\@\@\@$destination\@\@\@$enable\n";
         close (FILTER) or neomailerror("$lang_err{'couldnt_close'} .filter.book!");
      }
      
      ## remove .filter.check ##
      unlink("$folderdir/.filter.check");
   }
   print "Location: $scripturl&action=editfilter&sort=$sort&folder=$escapedfolder&firstmessage=$firstmessage&amp;message_id=$escapedmessageid\n\n";
}
################## END MODFILTER ###########################

###################### READPREFS #########################
sub readprefs {
   my ($key,$value);
   my %prefshash;
   if ( -f "$userprefsdir$UIDNM{$menu_id}/config" ) {
      open (CONFIG,"$userprefsdir$UIDNM{$menu_id}/config") or
         neomailerror("$lang_err{'couldnt_open'} config!");
      while (<CONFIG>) {
         ($key, $value) = split(/=/, $_);
         chomp($value);
         if ($key eq 'style') {
            $value =~ s/\.//g;  ## In case someone gets a bright idea...
         }
         $prefshash{"$key"} = $value;
      }
      close (CONFIG) or neomailerror("$lang_err{'couldnt_close'} config!");
   }
   if ( -f "$userprefsdir$UIDNM{$menu_id}/signature" ) {
      $prefshash{"signature"} = '';
      open (SIGNATURE, "$userprefsdir$UIDNM{$menu_id}/signature") or
         neomailerror("$lang_err{'couldnt_open'} signature!");
      while (<SIGNATURE>) {
         $prefshash{"signature"} .= $_;
      }
      close (SIGNATURE) or neomailerror("$lang_err{'couldnt_close'} signature!");
   }
   return \%prefshash;
}
##################### END READPREFS ######################

###################### READSTYLE #########################
sub readstyle {
   my ($key,$value);
   my $stylefile = $prefs{"style"} || 'Default';
   my %stylehash;
   unless ( -f "$stylesdir$stylefile") {
      $stylefile = 'Default';
   }
   open (STYLE,"$stylesdir$stylefile") or
      neomailerror("$lang_err{'couldnt_open'} $stylefile!");
   while (<STYLE>) {
      if (/###STARTSTYLESHEET###/) {
         $stylehash{"css"} = '';
         while (<STYLE>) {
            $stylehash{"css"} .= $_;
         }
      } else {
         ($key, $value) = split(/=/, $_);
         chomp($value);
         $stylehash{"$key"} = $value;
      }
   }
   close (STYLE) or neomailerror("$lang_err{'couldnt_close'} $stylefile!");
   return \%stylehash;
}
##################### END READSTYLE ######################

###################### SAVEPREFS #########################
sub saveprefs {
   unless ( -d "$userprefsdir$UIDNM{$menu_id}" ) {
      mkdir ("$userprefsdir$UIDNM{$menu_id}", 0700) or
         neomailerror("$lang_err{'cant_create_dir'}");
   }
   open (CONFIG,">$userprefsdir$UIDNM{$menu_id}/config") or
      neomailerror("$lang_err{'couldnt_open'} config!");
   foreach my $key (qw(realname domainname replyto sort headers style
                       numberofmessages language fromname)) {
      my $value = $DATA{"$key"} || '';
      $value =~ s/[\n|=|\/|\||\\|\`]//; # Strip out any sort of nastiness.
      if ($key eq 'language') {
         my $validlanguage=0;
         my $currlanguage;
         foreach $currlanguage (@availablelanguages) {
            if ($value eq $currlanguage) {
               print CONFIG "$key=$value\n";
               last;
            }
         }
      } elsif ($key eq 'fromname') {
         $value =~ s/\s+//g; # Spaces will just screw people up.
         print CONFIG "$key=$value\n";
      } else {
         print CONFIG "$key=$value\n";
      }
   }
   close (CONFIG) or neomailerror("$lang_err{'couldnt_close'} config!");
   open (SIGNATURE,">$userprefsdir$UIDNM{$menu_id}/signature") or
      neomailerror("$lang_err{'couldnt_open'} signature!");
   my $value = $DATA{"signature"} || '';
   if (length($value) > 500) {  # truncate signature to 500 chars
      $value = substr($value, 0, 500);
   }
   print SIGNATURE $value;
   close (SIGNATURE) or neomailerror("$lang_err{'couldnt_close'} signature!");
   printheader();

   my $html = '';
   my $temphtml;

   open (PREFSSAVEDTEMPLATE, "$neomaildir/templates/$lang/prefssaved.template") or
      neomailerror("$lang_err{'couldnt_open'} prefssaved.template!");
   while (<PREFSSAVEDTEMPLATE>) {
      $html .= $_;
   }
   close (PREFSSAVEDTEMPLATE);

   $html = applystyle($html);

   $temphtml = '<FORM METHOD="POST" action="'.$cgi_url.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="step" VALUE="mail">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="action" VALUE="displayheaders">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="sort" VALUE="'.$sort.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="firstmessage" VALUE="'.$firstmessage.'">';
   $temphtml .= '<INPUT TYPE="hidden" NAME="folder" VALUE="'.$folder.'">';
   $temphtml .= '<INPUT TYPE="submit" NAME="'.$lang_text{'continue'}.'" value="'.$lang_text{'continue'}.'"></form>';
   $html =~ s/\@\@\@CONTINUEBUTTON\@\@\@/$temphtml/;
   print $html;

   printfooter();
}
##################### END SAVEPREFS ######################

#################### ADDRESSBOOK #######################
sub addressbook {
   printheader();
   print "<center><h3>$lang_text{'abooktitle'}</h3></center>";
   my %addresses;
   my ($name, $email);
   my $field=$DATA{"field"};
   my $preexisting = $DATA{"preexisting"} || '';
   $preexisting =~ s/(\s+)?,(\s+)?/,/g;
   print '<FORM METHOD="POST" action="javascript:Update('."'".$field."'".')" name="addressbook">';
   print '<table border="0" align="center" width="90%" cellpadding="0" cellspacing="0">';
   print '<tr><td colspan="2" bgcolor='.$style{"titlebar"}.' align="left">';
   print '<font color='.$style{"titlebar_text"}.' face='.$style{"fontface"}.' size="3"><b>'.uc($lang_text{$field}).": $lang_text{'abook'}</b></font></td></tr>";

   if ( -f "$folderdir/.address.book" ) {
      open (ABOOK,"$folderdir/.address.book") or
         neomailerror("$lang_err{'couldnt_open'} .address.book!");
      while (<ABOOK>) {
         ($name, $email) = split(/:/, $_);
         chomp($email);
         $addresses{"$name"} = $email;
      }
      close (ABOOK) or neomailerror("$lang_err{'couldnt_close'} .address.book!");

      my $bgcolor = $style{"tablerow_dark"};
      foreach my $key (sort(keys %addresses)) {
         print "<tr><td bgcolor=$bgcolor width=\"20\"><input type=\"checkbox\" name=\"to\" value=\"".$addresses{$key}.'"';
         if ($preexisting =~ s/\Q$addresses{$key}\E,?//g) {
            print " checked";
         }
         print "></td><td width=\"100%\" bgcolor=$bgcolor>$key</td></tr>\n";
         if ($bgcolor eq $style{"tablerow_dark"}) {
            $bgcolor = $style{"tablerow_light"};
         } else {
            $bgcolor = $style{"tablerow_dark"};
         }
      }
   }
   print '</td></tr><tr><td align="center" colspan="2" bgcolor='.$style{"tablerow_dark"}.'>';
   print '<input type="hidden" name="remainingstr" value="'.$preexisting.'">';
   print '<input type="submit" name="mailto.x" value="OK"> &nbsp;&nbsp;';
   print '<input type="button" value="Cancel" onClick="window.close();">';
   print '</td></tr></table>';
   print '<script language="JavaScript">
      <!--
      function Update(whichfield)
      {
         var e2 = document.addressbook.remainingstr.value;
         for (var i = 0; i < document.addressbook.elements.length; i++)
         {
            var e = document.addressbook.elements[i];
            if (e.name == "to" && e.checked)
            {
               if (e2)
                  e2 += ",";
               e2 += e.value;
            }
         }
         if (whichfield == "to")
            window.opener.document.composeform.to.value = e2;
         else if (whichfield == "cc")
            window.opener.document.composeform.cc.value = e2;
         else
            window.opener.document.composeform.bcc.value = e2;
         window.close();
      }
      //-->
      </script>';
   print '</form></html>';
}
################## END ADDRESSBOOK #####################

1;